import logging, h5py, tables, gzip, time
import os, sys, shutil, psutil, numpy as np
import re, haplotyping, ahocorasick, pickle, signal
import haplotyping.index.database
import multiprocessing as mp
from queue import Empty

class Splits:
    
    """
    Internal use, get splitting k-mers from index
    """
    
    stepSizeStorage = 1000000
    
    def __init__(self, sortedIndexFile: str, h5file, filenameBase, debug=False, keepTemporaryFiles=False):
        
        """
        Internal use only: initialize
        """
        
        #logger
        self._logger = logging.getLogger(__name__)
        self._logger.info("parse k-mers to find right splitting k-mers")
        
        #set variables
        self.k = h5file["/config"].attrs["k"]
        self.automatonKmerSize = h5file["/config"].attrs["automatonKmerSize"]
        self.minimumFrequency = h5file["/config"].attrs["minimumCanonicalSplitFrequency"]
        self.h5file = h5file
        self.maximumNumber = 0
        self.debug = debug
        self.keepTemporaryFiles = keepTemporaryFiles
        
        
        #check existence group
        if not "/split" in h5file:
            h5file.create_group("/split")
        if not "/histogram" in h5file:
            h5file.create_group("/histogram")
        
        #create splitting k-mer dataset
        if "/split/ckmer" in h5file:
            self._logger.warning("canonical splitting k-mer dataset already in hdf5 storage")
        elif "/histogram/kmer" in h5file:
            self._logger.warning("histogram k-mer dataset already in hdf5 storage")
        elif "/histogram/ckmer" in h5file:
            self._logger.warning("histogram splitting k-mer dataset already in hdf5 storage")
        elif "/histogram/base" in h5file:
            self._logger.warning("histogram splitting k-mer bases dataset already in hdf5 storage")
        else:
        
            try:                                                
                pytablesFile = filenameBase+"_tmp_split.h5"
                if os.path.exists(pytablesFile):
                    os.remove(pytablesFile)
                self._logger.debug("store temporary in "+pytablesFile)    
                
                #create datasets
                with tables.open_file(pytablesFile, mode="w", 
                                      title="Temporary storage") as pytablesStorage:
                    #get k-mers
                    self._parseIndex(sortedIndexFile, pytablesStorage)
                    self._sort(pytablesStorage)
                    self._store(pytablesStorage)                    
                    #flush
                    self.h5file.flush()                    
            except Exception as e:
                self._logger.error("problem occurred while constructing splits: "+str(e))
            finally:
                try:
                    if not self.keepTemporaryFiles:
                        os.remove(pytablesFile)
                except:
                    self._logger.error("problem removing "+pytablesFile)    

#----------------
# Main functions
#----------------            
                    
    def _parseIndex(self, filename: str, pytablesStorage):
        
        
        def close_queue(queue_entry):
            time.sleep(0.1)
            queue_entry.close()
            queue_entry.join_thread()
            
        
        #create temporary storage
        tableCkmerDef = {
            "ckmer": tables.StringCol(self.k,pos=0),
            "type": tables.StringCol(1,pos=1),
            "number": tables.UInt64Col(pos=2),
        }
        tableDumpKmers = pytablesStorage.create_table(pytablesStorage.root,
                                    "dumpCkmer",tableCkmerDef, "Temporary to store dump canonical k-mers")
        
        #multiprocessing
        with mp.Manager() as manager:
            shutdown_event = mp.Event()
            qsize = 10000
            queue_splits = mp.Queue(qsize)
            totalNumberOfKmers = mp.Value("L",0)
            totalSumOfKmerFrequencies = mp.Value("L",0)
            minimumAllKmerFrequencies = mp.Value("L",2**32)
            maximumAllKmerFrequencies = mp.Value("L",0)
            histogramKmer = manager.dict()
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            process_list = mp.Process(target=haplotyping.index.splits.Splits.workerList, 
                                      args=(shutdown_event,filename,self.k,self.minimumFrequency,
                                            totalNumberOfKmers,totalSumOfKmerFrequencies,
                                            minimumAllKmerFrequencies,maximumAllKmerFrequencies,histogramKmer,
                                            queue_splits, ))
            signal.signal(signal.SIGINT, original_sigint_handler)

            try:
                #start
                process_list.start()
                #init
                rightSplitBases = 0
                rightSplitKmers = 0
                self.frequencyHistogram = {"kmer": {}, "ckmer": {}, "base": {}}
                #process queue
                while True:
                    try:
                        item = queue_splits.get(block=True, timeout=1)
                        if item==None:
                            break
                        else:
                            rightSplitBases+=1
                            for key,value in item[1].items():
                                kmer = item[0]+key
                                ckmer = haplotyping.General.canonical(kmer)
                                #assume right splitting if k-mer equals rc
                                #todo: can this (in theory) cause problematic situations?
                                ckmerType = "r" if kmer==ckmer else "l"
                                ckmerRow = tableDumpKmers.row            
                                ckmerRow["ckmer"] = ckmer
                                ckmerRow["type"] = ckmerType
                                ckmerRow["number"] = value
                                ckmerRow.append()
                                rightSplitKmers+=1 
                                self.frequencyHistogram["ckmer"][value] = (
                                    self.frequencyHistogram["ckmer"].get(value,0)+1)
                            if (rightSplitBases%1000000)==0:
                                self._logger.debug("processed {} bases and {} k-mers".format(
                                    rightSplitBases,rightSplitKmers))
                    except Empty:
                        continue
                tableDumpKmers.flush()
                self._logger.debug("found {} rightSplitBases and {} rightSplitKmers".format(
                    rightSplitBases,rightSplitKmers))
                self.h5file["/config/"].attrs["numberRightSplitBases"]=rightSplitBases
                self.h5file["/config/"].attrs["numberRightSplitKmers"]=rightSplitKmers

            except KeyboardInterrupt:
                self._logger.debug("caught KeyboardInterrupt")
                #shutdown directly
                shutdown_event.set()
                self._logger.debug("forced exit")
                sys.exit()
            finally:
                #close queus workers
                close_queue(queue_splits)
                #shutdown
                shutdown_event.set()
                #join
                process_list.join() 

            #store stats
            self.frequencyHistogram["kmer"] = dict(histogramKmer)
            self.h5file["/config/"].attrs["numberKmers"]=totalNumberOfKmers.value
            self.h5file["/config/"].attrs["totalKmerFrequencies"]=totalSumOfKmerFrequencies.value
            self.h5file["/config/"].attrs["minimumKmerFrequencies"]=minimumAllKmerFrequencies.value
            self.h5file["/config/"].attrs["maximumKmerFrequencies"]=maximumAllKmerFrequencies.value
            
    def workerList(shutdown_event,filename,k,minimumFrequency,totalNumberOfKmers,totalSumOfKmerFrequencies,
                                        minimumAllKmerFrequencies,maximumAllKmerFrequencies,histogramKmer,
                                        queue_splits):
        logger = logging.getLogger("{}.worker.list".format(__name__))
        try:
            with gzip.open(filename, "rt") as f: 
                logger.debug("parse sorted list to detect right splitting k-mers")
                previousBase = ""
                previousBranch = ""
                previousKmer = ""
                previousNumber = 0                 
                #administration
                rightSplitBases = 0
                rightSplitKmers = 0
                stored={}
                errorNumber = 0
                _totalNumberOfKmers=0
                _totalSumOfKmerFrequencies=0
                _minimumAllKmerFrequencies=2**32
                _maximumAllKmerFrequencies=0
                _histogramKmer = {}
                while (row := f.readline()):
                    line = row.strip().split("\t") 
                    currentBase = line[0][:-1]
                    currentBranch = line[0][-1]
                    currentKmer = line[0]
                    _totalNumberOfKmers+=1
                    if (_totalNumberOfKmers%10000000)==0 and (_totalNumberOfKmers>0):
                        logger.debug("processed {} k-mers".format(_totalNumberOfKmers))
                    currentNumber = int(line[1])
                    _totalSumOfKmerFrequencies += currentNumber
                    _minimumAllKmerFrequencies = min(_minimumAllKmerFrequencies,currentNumber)
                    _maximumAllKmerFrequencies = max(_maximumAllKmerFrequencies,currentNumber)
                    _histogramKmer[currentNumber] = (_histogramKmer.get(currentNumber,0)+1)
                    if currentNumber < minimumFrequency:
                        pass
                    else:
                        if currentBase==previousBase:
                            stored[previousBranch]=previousNumber
                            stored[currentBranch]=currentNumber
                        else:
                            if len(stored)>1:
                                queue_splits.put((previousBase,stored,))
                                stored={}                                        
                        previousBase = currentBase
                        previousBranch = currentBranch
                        previousKmer = currentKmer
                        previousNumber = currentNumber
                if len(stored)>1:
                    queue_splits.put((previousBase,stored,)) 
                #store stats
                totalNumberOfKmers.value=_totalNumberOfKmers
                totalSumOfKmerFrequencies.value=_totalSumOfKmerFrequencies
                minimumAllKmerFrequencies.value=_minimumAllKmerFrequencies
                maximumAllKmerFrequencies.value=_maximumAllKmerFrequencies
                for key in _histogramKmer.keys():
                    histogramKmer[key] = _histogramKmer[key]
                #close queue
                queue_splits.put(None)
                #warning
                if errorNumber>0:
                    logger.warning("skipped {} items in sorted list".format(errorNumber))
                #stats
                minimumAllKmerFrequencies.value=min(minimumAllKmerFrequencies.value,maximumAllKmerFrequencies.value)
                logger.debug("checked {} k-mers with {} total frequency".format(
                    totalNumberOfKmers.value,totalSumOfKmerFrequencies.value))
                logger.debug("frequency k-mers between {} and {}".format(
                    minimumAllKmerFrequencies.value,maximumAllKmerFrequencies.value))
        except Exception as ex:
            logger.error("problem with sorted list: "+str(ex))
    

    def _sort(self, pytablesStorage):                        
        
        def saveToSortedCkmerStorage(ckmer,ckmerType,number,numberOfCkmers):
            ckmerRow = tableSortedKmers.row            
            ckmerRow["ckmer"] = ckmer
            ckmerRow["type"] = ckmerType
            ckmerRow["number"] = number
            ckmerRow.append()
            if ckmerType=="r" or ckmerType=="b":
                saveToDumpBaseStorage(ckmer[:-1],ckmer[-1],number,numberOfCkmers)
            if ckmerType=="l" or ckmerType=="b":
                rkmer=haplotyping.General.reverse_complement(ckmer)
                saveToDumpBaseStorage(rkmer[:-1],rkmer[-1],number,numberOfCkmers)
            return numberOfCkmers+1
        
        def saveToSortedBaseStorage(base,branches,number,numberOfBases):
            baseRow = tableSortedBases.row    
            baseRow["base"] = base
            baseRow["number"] = number
            self.frequencyHistogram["base"][number] = (
                                self.frequencyHistogram["base"].get(number,0)+1)
            for letter in branches.keys():
                baseRow["branches/"+str(letter)+"/number"] = branches[letter][0]
                baseRow["branches/"+str(letter)+"/ckmerLink"] = branches[letter][1]
                ckmerRow = tableSortedKmers[branches[letter][1]]
                ckmer = ckmerRow["ckmer"].decode()
                rkmer=haplotyping.General.reverse_complement(ckmer)
                if ckmer == base+letter:
                    baseReferences[branches[letter][1]][1] = numberOfBases
                if rkmer == base+letter:
                    baseReferences[branches[letter][1]][0] = numberOfBases
            baseRow.append()
            return numberOfBases+1
            
        def saveToDumpBaseStorage(base,branch,number,ckmerLink):   
            baseRow = tableDumpRightSplitBases.row
            baseRow["base"] = base
            baseRow["branch"] = branch
            baseRow["number"] = number
            baseRow["ckmerLink"] = ckmerLink
            baseRow.append()
            
        tableDumpKmers = pytablesStorage.root.dumpCkmer
        
        #create sorted and grouped storage
        tableDumpKmers.flush()                
        numberOfSplittingKmers=tableDumpKmers.shape[0]
        if numberOfSplittingKmers>0:
            
            #create temporary storage
            tableCkmerDef = {
                "ckmer": tables.StringCol(self.k,pos=0),
                "type": tables.StringCol(1,pos=1),
                "rightSplitBaseLink": {
                    "leftSplit": haplotyping.index.Database.getTablesUint(numberOfSplittingKmers,0),
                    "rightSplit": haplotyping.index.Database.getTablesUint(numberOfSplittingKmers,1),
                },
                "number": tables.UInt32Col(pos=3,dflt=0),
            }
            tableSortedKmers = pytablesStorage.create_table(pytablesStorage.root,
                                        "sortedCkmer",tableCkmerDef, "Temporary to store sorted canonical k-mers",
                                        expectedrows=numberOfSplittingKmers)  
            
            #create dump table for bases
            tableDumpBaseDef = {
                "base": tables.StringCol(self.k-1,pos=0),
                "branch": tables.StringCol(1,pos=1),
                "number": tables.UInt32Col(pos=2,dflt=0),
                "ckmerLink": haplotyping.index.Database.getTablesUint(numberOfSplittingKmers,3),
            }
            tableDumpRightSplitBases = pytablesStorage.create_table(pytablesStorage.root,
                                                                   "dumpRightSplitBase",tableDumpBaseDef, 
                                                                   "Temporary to store dump bases",
                                                                       expectedrows=numberOfSplittingKmers)
            
            #store sorted k-mers
            self._logger.debug("sort and group "+str(numberOfSplittingKmers)+" splitting k-mers")
            tableDumpKmers.cols.ckmer.create_csindex()
            tableDumpKmers.flush()
            self._logger.debug("created index to sort k-mers")
            previousCkmer=None
            previousType=None
            previousN=None
            self.maximumNumber = 0
            numberOfCkmers = 0
            for row in tableDumpKmers.itersorted("ckmer",checkCSI=True):
                currentCkmer=row["ckmer"]
                currentType=row["type"]
                currentN=row["number"]
                if previousCkmer and not previousCkmer==currentCkmer:
                    numberOfCkmers = saveToSortedCkmerStorage(previousCkmer.decode(),previousType,
                                                         previousN,numberOfCkmers)
                    self.maximumNumber=max(self.maximumNumber,previousN)
                    previousType=currentType.decode()
                elif previousCkmer and not previousType==currentType:
                    previousType="b"
                else:
                    previousType=currentType.decode()
                previousCkmer=currentCkmer
                previousN=currentN
                if (numberOfCkmers%1000000)==0 and (numberOfCkmers>0):
                    self._logger.debug("processed {} k-mers".format(numberOfCkmers))
            if previousCkmer:        
                numberOfCkmers = saveToSortedCkmerStorage(previousCkmer.decode(),previousType,
                                                     previousN,numberOfCkmers) 
                self.maximumNumber=max(self.maximumNumber,previousN)
            self._logger.debug("in total processed {} k-mers".format(numberOfCkmers))
            
            tableDumpRightSplitBases.flush()                
            numberOfBases=tableDumpRightSplitBases.shape[0]
                
            #store sorted bases
            tableSortedKmers.flush()
            numberOfKmers=tableSortedKmers.shape[0]
            tableBaseDef = {
                "base": tables.StringCol(self.k-1,pos=0),
                "number": tables.UInt32Col(pos=1,dflt=0),    
                "branches": {},                            
            }
            for letter in haplotyping.index.Database.letters:
                tableBaseDef["branches"][letter] = {
                    "number": tables.UInt32Col(pos=0,dflt=0),
                    "ckmerLink": haplotyping.index.Database.getTablesUint(numberOfKmers,1),
                }
            tableSortedBases = pytablesStorage.create_table(pytablesStorage.root,
                                                                   "sortedBase",tableBaseDef, 
                                                                   "Temporary to store dump bases",
                                                                       expectedrows=numberOfKmers)
            
            #store sorted bases
            self._logger.debug("sort and group {} bases".format(numberOfBases))
            tableDumpRightSplitBases.cols.base.create_csindex()
            tableDumpRightSplitBases.flush()
            self._logger.debug("created index to sort bases")
            previousBase=None
            previousBranches={}
            previousNumber=0
            numberOfBases = 0
            #reserve memory for links in ckmer table
            ckmerLinkType = np.dtype(haplotyping.index.Database.getUint(numberOfKmers)).type
            baseReferences = np.full((numberOfKmers,2), 0, dtype=ckmerLinkType)
            for row in tableDumpRightSplitBases.itersorted("base",checkCSI=True):
                currentBase=row[0]
                currentBranch=row[1]
                currentNumber=row[2]
                currentCkmerLink=row[3]
                if previousBase and not previousBase==currentBase:
                    numberOfBases = saveToSortedBaseStorage(previousBase.decode(),previousBranches,
                                                            previousNumber,numberOfBases)
                    previousBase=None
                    previousBranches={}
                    previousNumber=0
                    if (numberOfBases%1000000)==0 and (numberOfBases>0):
                        self._logger.debug("processed {} bases".format(numberOfBases))
                previousBase=currentBase
                previousBranches[currentBranch.decode()] = (currentNumber, currentCkmerLink,)
                previousNumber+=currentNumber                    
            if previousBase:
                    numberOfBases = saveToSortedBaseStorage(previousBase.decode(),previousBranches,
                                                            previousNumber,numberOfBases)  
            self._logger.debug("in total processed {} bases".format(numberOfBases))
            #update ckmers
            tableSortedKmers.modify_columns(columns=baseReferences,
                             names=["rightSplitBaseLink/leftSplit","rightSplitBaseLink/rightSplit"])   
                
        else:
            self._logger.warning("no splitting k-mers to sort and group")
            
    def _store(self, pytablesStorage):
        canonicalSplitKmers = 0
        canonicalSplitKmersLeft = 0
        canonicalSplitKmersRight = 0
        canonicalSplitKmersBoth = 0
        tableSortedKmers = pytablesStorage.root.sortedCkmer
        tableSortedBases = pytablesStorage.root.sortedBase
        #create storage    
        tableSortedKmers.flush()
        tableSortedBases.flush()
        numberOfKmers=tableSortedKmers.shape[0]
        numberOfBases=tableSortedBases.shape[0]
        # CKMER STORAGE - don't make the structure unnecessary big
        dtypeCkmerList=[("ckmer","S"+str(self.k)),
                   ("type","S1"),
                   ("number",haplotyping.index.Database.getUint(self.maximumNumber)),
                   ("rightSplitBaseLink",[("leftSplit",haplotyping.index.Database.getUint(numberOfBases)),
                                          ("rightSplit",haplotyping.index.Database.getUint(numberOfBases))]),
                   ("direct",[("link", haplotyping.index.Database.getUint(4*numberOfKmers)),
                              ("left",
                               [("distinct","uint8"),
                                ("number",haplotyping.index.Database.getUint(self.maximumNumber))]),
                              ("right",
                               [("distinct","uint8"),
                                ("number",haplotyping.index.Database.getUint(self.maximumNumber))])
                             ]),
                   ("partition",haplotyping.index.Database.getUint(numberOfKmers)),
                   ("cycle",[("number",haplotyping.index.Database.getUint(self.maximumNumber))]),
                   ("reversal",[("number",haplotyping.index.Database.getUint(self.maximumNumber))]),
                   ("paired",[("link",haplotyping.index.Database.getUint(numberOfKmers)),
                              ("number",haplotyping.index.Database.getUint(self.maximumNumber))])]
        dtCkmer=np.dtype(dtypeCkmerList)
        dsCkmer=self.h5file["/split/"].create_dataset("ckmer",(numberOfKmers,), dtype=dtCkmer, chunks=None, 
                                                      compression="gzip", compression_opts=9)
        self._logger.info("store {} splitting k-mers".format(numberOfKmers))
        #add stored and grouped kmers to the final unchunked storage
        for i in range(0,numberOfKmers,Splits.stepSizeStorage):
            stepData = tableSortedKmers[i:i+Splits.stepSizeStorage]
            dsCkmer[i:i+Splits.stepSizeStorage] = stepData
            for row in stepData:
                canonicalSplitKmers+=1
                if row[1].decode()=="l":
                    canonicalSplitKmersLeft+=1
                elif row[1].decode()=="b":
                    canonicalSplitKmersBoth+=1
                elif row[1].decode()=="r":
                    canonicalSplitKmersRight+=1
        dsCkmer.flush()
        # BASE STORAGE - don't make the structure unnecessary big
        dtypeBaseList=[("base","S"+str(self.k-1)),
                   ("number",haplotyping.index.Database.getUint(self.maximumNumber)),
                   ("branches",[])]
        for i in range(len(haplotyping.index.Database.letters)):
            letter = haplotyping.index.Database.letters[i]
            dtypeBaseList[2][1].append((letter,[("number",haplotyping.index.Database.getUint(self.maximumNumber)),
                                                ("ckmerLink",numberOfKmers)]))
        dtBase=np.dtype(dtypeBaseList)
        dsBase=self.h5file["/split/"].create_dataset("base",(numberOfBases,), dtype=dtBase, chunks=None,
                                                     compression="gzip", compression_opts=9)
        self._logger.info("store {} bases".format(numberOfBases))
        #add stored and grouped bases to the final unchunked storage
        for i in range(0,numberOfBases,Splits.stepSizeStorage):
            stepData = tableSortedBases[i:i+Splits.stepSizeStorage]
            dsBase[i:i+Splits.stepSizeStorage] = stepData
        # HISTOGRAM K-MER STORAGE - don't make the structure unnecessary big
        maximumFrequency = max(self.frequencyHistogram["kmer"].keys())
        maximumNumber = max(self.frequencyHistogram["kmer"].values())
        dtypeFrequencyHistogramKmerList=[
                    ("frequency",haplotyping.index.Database.getUint(maximumFrequency)),
                    ("number",haplotyping.index.Database.getUint(maximumNumber)),]
        dtFrequencyHistogramKmer=np.dtype(dtypeFrequencyHistogramKmerList)
        dsFrequencyHistogramKmer=self.h5file["/histogram/"].create_dataset("kmer",(len(self.frequencyHistogram["kmer"]),), 
                                                                  dtype=dtFrequencyHistogramKmer, chunks=None, 
                                                                  compression="gzip", compression_opts=9)
        self._logger.info("store {} entries k-mer histogram".format(len(self.frequencyHistogram["kmer"])))
        #store histogram
        dsFrequencyHistogramKmer[0:len(self.frequencyHistogram["kmer"])] = list(
            sorted(self.frequencyHistogram["kmer"].items()))
        # HISTOGRAM SPLITTING K-MER STORAGE - don't make the structure unnecessary big
        maximumFrequency = max(self.frequencyHistogram["ckmer"].keys())
        maximumNumber = max(self.frequencyHistogram["ckmer"].values())
        dtypeFrequencyHistogramCkmerList=[
                    ("frequency",haplotyping.index.Database.getUint(maximumFrequency)),
                    ("number",haplotyping.index.Database.getUint(maximumNumber)),]
        dtFrequencyHistogramCkmer=np.dtype(dtypeFrequencyHistogramCkmerList)
        dsFrequencyHistogramCkmer=self.h5file["/histogram/"].create_dataset("ckmer",(len(self.frequencyHistogram["ckmer"]),), 
                                                                  dtype=dtFrequencyHistogramCkmer, chunks=None, 
                                                                  compression="gzip", compression_opts=9)
        self._logger.info("store {} entries splitting k-mer histogram".format(len(self.frequencyHistogram["ckmer"])))
        #store histogram
        dsFrequencyHistogramCkmer[0:len(self.frequencyHistogram["ckmer"])] = list(
            sorted(self.frequencyHistogram["ckmer"].items()))
        # HISTOGRAM SPLITTING K-MER BASES STORAGE - don't make the structure unnecessary big
        maximumFrequency = max(self.frequencyHistogram["base"].keys())
        maximumNumber = max(self.frequencyHistogram["base"].values())
        dtypeFrequencyHistogramBaseList=[
                    ("frequency",haplotyping.index.Database.getUint(maximumFrequency)),
                    ("number",haplotyping.index.Database.getUint(maximumNumber)),]
        dtFrequencyHistogramBase=np.dtype(dtypeFrequencyHistogramBaseList)
        dsFrequencyHistogramBase=self.h5file["/histogram/"].create_dataset("base",(len(self.frequencyHistogram["base"]),), 
                                                                  dtype=dtFrequencyHistogramBase, chunks=None, 
                                                                  compression="gzip", compression_opts=9)
        self._logger.info("store {} entries splitting k-mer bases histogram".format(len(self.frequencyHistogram["base"])))
        #store histogram
        dsFrequencyHistogramBase[0:len(self.frequencyHistogram["base"])] = list(
            sorted(self.frequencyHistogram["base"].items()))
        #store the stats
        self._logger.debug("found {} canonicalSplitKmersLeft".format(canonicalSplitKmersLeft))
        self._logger.debug("found {} canonicalSplitKmersRight".format(canonicalSplitKmersRight))
        self._logger.debug("found {} canonicalSplitKmersBoth".format(canonicalSplitKmersBoth))
        self.h5file["/config/"].attrs["maximumCanonicalSplitFrequency"]=self.maximumNumber
        self.h5file["/config/"].attrs["numberCanonicalSplit"]=canonicalSplitKmers
        self.h5file["/config/"].attrs["numberCanonicalSplitLeft"]=canonicalSplitKmersLeft
        self.h5file["/config/"].attrs["numberCanonicalSplitBoth"]=canonicalSplitKmersBoth
        self.h5file["/config/"].attrs["numberCanonicalSplitRight"]=canonicalSplitKmersRight 
        
#-------------------------------------------------
# COMPUTATION AUTOMATON AND INDEX SPLITTING K-MER
#-------------------------------------------------

    def deleteAutomatonWithIndex(filenameBase, k):
        indexFile = "{}_{}.index.splits".format(filenameBase,k)
        automatonFile = "{}_{}.automaton.splits".format(filenameBase,k)
        automatonStatsFile = "{}_{}.automaton.splits.stats".format(filenameBase,k)
        if os.path.exists(indexFile):
            os.remove(indexFile)
        if os.path.exists(automatonFile):
            os.remove(automatonFile)
        if os.path.exists(automatonStatsFile):
            os.remove(automatonStatsFile)
        
    def createAutomatonWithIndex(h5file, filenameBase, k):
        k = min(h5file["/config"].attrs["k"],k)
        logger = logging.getLogger(__name__)
        logger.info("create automaton with k' = {}".format(k))
        indexFile = "{}_{}.index.splits".format(filenameBase,k)
        automatonFile = "{}_{}.automaton.splits".format(filenameBase,k)
        automatonStatsFile = "{}_{}.automaton.splits.stats".format(filenameBase,k)
        if os.path.exists(indexFile) and os.path.exists(automatonFile)and os.path.exists(automatonStatsFile):
            logger.debug("detected previously generated automaton and index")
            with open(automatonStatsFile, "rb") as f:
                stats = pickle.load(f)
            logger.debug("automaton with {} words, size {} MB".format(stats["words_count"],
                                                                          round(stats["real_size"]/1048576)))
            logger.debug("associated index filesize {} MB".format(round(os.stat(indexFile).st_size/1048576)))
            return (max(stats["real_size"],stats["total_size"]),indexFile, automatonFile)
        else:
            logger.debug("generate automaton and index")
            tmpAutomatonFile = "{}.tmp".format(automatonFile)
            tmpIndexFile = "{}.tmp".format(indexFile)
            try:
                process = psutil.Process(os.getpid())
                memoryBefore = process.memory_info().rss
                automatonSplits = ahocorasick.Automaton()
                numberOfKmers = h5file["/split/ckmer"].shape[0]
                kmers = h5file["/split/ckmer"]
                with open(tmpIndexFile, "w") as f:
                    for i in range(0,numberOfKmers,Splits.stepSizeStorage):
                        kmerSubset = kmers[i:i+Splits.stepSizeStorage]
                        for j in range(len(kmerSubset)):
                            id = i + j
                            row = kmerSubset[j]                
                            kmer = row[0].decode()
                            #store in index
                            f.write(kmer)
                            #build automaton for k'-mers with k'<=k describing:
                            #- number of canonical k-mers to check starting from the provided index-location
                            #- index-location of the first matching canonical k-mer
                            rkmer = haplotyping.General.reverse_complement(kmer[-k:])
                            kmer = kmer[:k]
                            if (not automatonSplits.exists(kmer)):
                                automatonSplits.add_word(kmer,(1,id))
                            else:
                                value=list(automatonSplits.get(kmer))
                                if value[0]==0:
                                    value[1]=id
                                    value[0]=1
                                else:
                                    value[0]=id-value[1]+1
                                automatonSplits.add_word(kmer,tuple(value))
                            if not kmer==rkmer:
                                if (not automatonSplits.exists(rkmer)):
                                    automatonSplits.add_word(rkmer,(0,id))
                #create and store automaton
                automatonSplits.make_automaton()
                memoryAfter = process.memory_info().rss
                stats = automatonSplits.get_stats()
                stats["real_size"] = memoryAfter - memoryBefore
                logger.debug("automaton with {} words, size {} MB".format(stats["words_count"],
                                                                          round(stats["real_size"]/1048576)))
                logger.debug("associated index filesize {} MB".format(round(os.stat(tmpIndexFile).st_size/1048576)))
                automatonSplits.save(tmpAutomatonFile, pickle.dumps)
                #clear
                automatonSplits.clear()
                del automatonSplits
                #copy all
                shutil.copyfile(tmpAutomatonFile, automatonFile)
                shutil.copyfile(tmpIndexFile, indexFile)
                #also store stats
                with open(automatonStatsFile, "wb") as f:
                    pickle.dump(stats, f)
            except Exception as ex:
                logger.debug("problem while creating automaton: {}".format(ex))
            finally:
                if os.path.exists(tmpIndexFile):
                    os.remove(tmpIndexFile)
                if os.path.exists(tmpAutomatonFile):
                    os.remove(tmpAutomatonFile)
                    
                
        return (max(stats["real_size"],stats["total_size"]),indexFile, automatonFile)
        
   

    


import logging,requests
import haplotyping.graph
import haplotyping.api
import haplotyping.graph.baseGraph as baseGraph
import pandas as pd
    
class APIGraph(baseGraph.Graph):
    """basic or minimal version of the De Bruijn graph using the dataset 
        defined by uid via API"""
    
    def __init__(self, url: str, uid: str, username: str = None, password: str = None):
        """use the API at provided url"""
        #call parent constructor
        super(haplotyping.graph.api.APIGraph, self).__init__()
        #logger
        self._api_logger = logging.getLogger(__name__)
        #store call parameters
        self._baseUrl : str = str(url)
        self._uid : str = str(uid)
        self._varietyUid : str = None
        self._username : str = username
        self._password : str = password
        #initialise
        self._variety : str = None
        self._collection : str = None
        self._datasetFrequencyCkmers = set()
        self._datasetFrequencies = {}
        self._datasetVarieties = {}
        self._datasetUids = set()
        #report to logger
        self._api_logger.debug("using dataset {}".format(self._uid))
        self._api_logger.debug("using API at {}".format(self._baseUrl))
        #api
        self._api = haplotyping.api.API(url, username=username, password=password)        
        #checks
        self._api_initial_checks()
        
    def _api_initial_checks(self):
        """
        get dataset and check for k-mer and split
        """
        self._api_logger.debug("get dataset information from API")        
        #get dataset info
        try:
            dataset = self._api.getDatasetById(self._uid)
            if not (dataset["type"]=="kmer" or dataset["type"]=="split"):
                self._api_logger.error("dataset {} has no k-mer database".format(self._uid))
            if not (dataset["type"]=="split"):
                self._api_logger.error("dataset {} has no splitting k-mer database".format(self._uid))   
            if dataset["variety"]:
                self._variety = str(dataset["variety"]["name"])
                self._varietyUid = str(dataset["variety"]["uid"])
                self._api_logger.debug("set variety to '{}'".format(self._variety))
            if dataset["collection"]:
                self._collection = str(dataset["collection"]["name"])
                self._api_logger.debug("set collection to '{}'".format(self._collection))
            if self._variety:
                self._name = "'{}'".format(self._variety)
                if self._collection:
                    self._name = "{} from '{}' collection".format(self._name,self._collection)
            #get split info
            try:
                data = self._api.getSplitInfo(self._uid)
                self._k = int(data["k"])
                self._api_logger.debug("set k-mer size to {}".format(self._k))
            except:
                self._api_logger.error("k-mer database {} not found".format(self._uid)) 
        except:
            self._api_logger.error("dataset {} not found".format(self._uid))
                    
    def getDatasetVarieties(self):
        return self._datasetVarieties
    
    def getDatasetFrequencies(self, datasetUids: list = None):
        #compute dataset uids
        if datasetUids==None:
            datasetUids = set()
            start = 0
            number = 1000
            while True:
                response = requests.get("{}dataset?hasVariety=true&dataType=kmer&start={}&number={}".format(
                    self._baseUrl, start, number))
                if response.ok:
                    data = response.json()
                    total = data.get("total",0)
                    for item in data["list"]:
                        datasetUids.add(item["uid"])
                    if start+number>=total:
                        break
                    else:
                        start+=number
                else:
                    self._api_logger.error("request to {} didn't succeed".format(self._baseUrl))            
        candidateKmers = [k[0] for k in self.getCandidates()] 
        recompute = False
        if len(set(datasetUids).difference(self._datasetUids))>0:
            recompute = True
        elif len(set(candidateKmers).difference(self._datasetFrequencyCkmers))>0:
            recompute = True
        if recompute:
            self._api_logger.debug("recompute frequency for {} k-mers in {} datasets".format(
                len(candidateKmers),len(datasetUids)))
            self._datasetFrequencyCkmers = set(candidateKmers)
            self._datasetFrequencies = {}
            self._datasetVarieties = {}
            self._datasetUids = set()
            #get dataset info
            fullRequest = "{}dataset/".format(self._baseUrl)
            response = requests.post(fullRequest, json = {"uids": sorted(list(datasetUids))}, 
                                     auth=self._apiAuth, headers=self._apiHeaders)
            if response.ok:
                data = response.json()
                result = {}
                items = data.get("list",[])
                for item in items:
                    dsUid = item["uid"]
                    if "variety" in item.keys():
                        self._datasetVarieties[dsUid] = item["variety"]
                    
            else:
                self._api_logger.error("request to {} didn't succeed".format(self._baseUrl))
            #get frequencies
            for dsUid in datasetUids:                
                fullRequest = "{}kmer/{}".format(self._baseUrl, dsUid)        
                response = requests.post(fullRequest, json={"kmers": sorted(candidateKmers),"mismatches": 0}, 
                                             auth=self._apiAuth, headers=self._apiHeaders)
                if response.ok:
                    data = response.json()
                    self._datasetFrequencies[dsUid] = data.get("kmers",{})
                    for k in candidateKmers:
                        if not k in self._datasetFrequencies[dsUid].keys():
                            self._datasetFrequencies[dsUid][k] = 0
                else:
                    self._api_logger.error("request to {} didn't succeed".format(self._baseUrl))
            #register datasets
            self._datasetUids.update(datasetUids)
        #return result
        return pd.DataFrame(self._datasetFrequencies).transpose()





from flask import Response, request, abort
from flask_restx import Namespace, Resource, fields
import json, haplotyping, sqlite3, re

namespace = Namespace("variety", description="All available varieties", path="/variety")
parser = namespace.parser()

yearPattern = re.compile(r"^[\d]{4}$")
yearMaxPattern = re.compile(r"^\<[\d]{4}$")
yearMinPattern = re.compile(r"^\>[\d]{4}$")
yearRangePattern = re.compile(r"^[\d]{4}\-[\d]{4}$")

def _make_bool(text):
    return str(text).lower() in ("yes", "true", "t", "1")

def get_variety_datasets(uid, collection, dataset, db_connection):
    condition_variables = [uid]
    condition_sql = """`dataset`.`variety` = ? 
                        AND (NOT `dataset`.`uid` IS NULL) AND (NOT `dataset`.`type` IS NULL)"""
    if not collection==None:
        collection_list = collection.split(",")
        condition_sql = condition_sql + " AND (`collection`.`uid` IN ("+",".join(["?"]*len(collection_list))+"))"
        condition_variables.extend(collection_list)
    if not dataset==None:
        if dataset=="any":
            condition_sql = condition_sql + " AND NOT (`dataset`.`id` IS NULL)"
        elif dataset=="none":
            condition_sql = condition_sql + " AND (`dataset`.`id` IS NULL)"  
        elif dataset=="kmer":
            condition_sql = condition_sql + " AND ((`dataset`.`type` IS 'kmer') OR (`dataset`.`type` IS 'split'))"  
        elif dataset=="split":
            condition_sql = condition_sql + " AND (`dataset`.`type` IS 'split')" 
        elif dataset=="marker":
            condition_sql = condition_sql + " AND (`dataset`.`type` IS 'marker')" 
        else:
            abort(422, "incorrect dataset condition "+str(dataset))
    cursor = db_connection.cursor()
    cursor.execute("""SELECT `dataset`.`uid`, 
                           `dataset`.`type`, 
                           `dataset`.`subtype`, 
                           `dataset`.`version`, 
                           `collection`.`uid` AS `collection_uid`, 
                           `collection`.`name` AS `collection_name`, 
                           `collection`.`type` AS `collection_type`, 
                           `collection`.`experiment` AS `collection_experiment` 
                        FROM `dataset` 
                        LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` 
                        WHERE """+condition_sql+""" ORDER BY `collection`.`name`, `dataset`.`uid`""", 
                   tuple(condition_variables)) 
    return [dict(row) for row in cursor.fetchall()]  

def parents_tree(id, parentData, collection, dataType, db_connection):
    parents = []   
    for item in parentData:
        if item["offspring"]==id:
            if item["ancestor"]==None:
                parent = {"relation": item["type"],
                          "parents": parents_tree(item["id"],parentData, collection, dataType, db_connection)}
                parents.append(parent)
            else:
                parent = {"uid": item["ancestor"],
                          "name": item["name"],
                          "relation": item["type"]}
                parent = append_variety_datasets(parent, collection, dataType, db_connection)
                parents.append(parent)
    return parents

def offspring_list(offspringData):
    offspring = []
    for item in offspringData:
        offspring_item = {
            "uid": item["variety"],
            "name": item["name"],
            "relation": "unknown" if item["type"]==None else item["type"],
            "direct": True if item["offspring"]==None else False
        }        
        offspring.append(offspring_item)
    return offspring
        
def get_variety_parents(uid, collection, dataType, db_connection):
    cursor = db_connection.cursor()
    cursor.execute("""SELECT `variety_ancestor`.*, `variety`.`name` FROM `variety_ancestor` 
                    LEFT JOIN `variety` ON `variety_ancestor`.`ancestor` = `variety`.`uid` 
                    WHERE `variety_ancestor`.`variety` = ? 
                    ORDER BY `variety_ancestor`.`type` DESC""",(uid,)) 
    parentData = [dict(row) for row in cursor.fetchall()]
    return parents_tree(None,parentData, collection, dataType, db_connection)

def get_variety_offspring(uid, collection, dataType, db_connection):
    cursor = db_connection.cursor()
    cursor.execute("""SELECT `variety_ancestor`.*, `variety`.`name` FROM `variety_ancestor` 
                    LEFT JOIN `variety` ON `variety_ancestor`.`variety` = `variety`.`uid` 
                    WHERE `variety_ancestor`.`ancestor` = ? 
                    ORDER BY `variety`.`name`, `variety`.`uid`""",(uid,)) 
    offspringData = [dict(row) for row in cursor.fetchall()]
    offspringList = offspring_list(offspringData)
    for j in range(len(offspringList)):
        offspringList[j] = append_variety_datasets(offspringList[j], collection, dataType, db_connection)
    return offspringList

def year_description(year_min,year_max):
    if year_min==None and year_max==None:
        return None
    elif year_min==None:
        return "<"+str(year_max+1)
    elif year_max==None:
        return ">"+str(year_min-1)
    elif year_min==year_max:
        return str(year_min)
    else:
        return str(year_min)+"-"+str(year_max)
    
def append_variety_datasets(item, collection, dataType, db_connection):
    item["datasets"] = get_variety_datasets(item["uid"], collection, dataType, db_connection)
    for i in range(len(item["datasets"])):
        item["datasets"][i]["collection"] = {
            "uid": item["datasets"][i]["collection_uid"],
            "name": item["datasets"][i]["collection_name"],
            "type": item["datasets"][i]["collection_type"],
            "experiment": item["datasets"][i]["collection_experiment"]
        }
        del item["datasets"][i]["collection_uid"]
        del item["datasets"][i]["collection_name"]
        del item["datasets"][i]["collection_type"]
        del item["datasets"][i]["collection_experiment"]
    return item
    
def adjust_variety_response(item, collection, dataType, db_connection):
    #origin
    if item["origin"]:
        item["origin"] = {"uid": item["origin"], "country": item["country"]}
    else:
        del item["origin"]
    del item["country"]
    #year
    if item["year_min"] or item["year_max"]:
        item["year"] = {"description": year_description(item["year_min"],item["year_max"]), 
                        "min": item["year_min"], 
                        "max": item["year_max"]}
    else:
        del item["year"]
    del item["year_min"]
    del item["year_max"]
    if "synonyms" in item and not item["synonyms"]:
        del item["synonyms"]
    #datasets
    item = append_variety_datasets(item, collection, dataType, db_connection)
    #parents with datasets
    item["parents"] = get_variety_parents(item["uid"], collection, dataType, db_connection)
    #offspring with datasets
    item["offspring"] = get_variety_offspring(item["uid"], collection, dataType, db_connection)    
    #finished
    return item
        
@namespace.route("/")
class VarietyList(Resource):
    
    variety_list = parser.copy()
    variety_list.add_argument("start", type=int, required=False, location="args", 
                              help="paging")       
    variety_list.add_argument("number", type=int, required=False, location="args", 
                              help="paging")
    variety_list.add_argument("name", type=str, required=False, location="args", 
                              help="name of variety")
    variety_list.add_argument("nameContains", type=str, required=False, location="args", 
                              help="name of variety contains (only if no name condition)")
    variety_list.add_argument("origin", type=str, required=False, location="args", 
                              help="comma separated list of country codes")
    variety_list.add_argument("year", type=str, required=False, location="args", 
                              help="year of variety (e.g. '1995', '<1995', '>1995', '1990-1995')")
    variety_list.add_argument("collection", type=str, required=False, location="args", 
                              help="variety has dataset from comma separated list of collection uids")
    variety_list.add_argument("dataType", type=str, required=False, location="args", 
                              help="variety has dataset (of specific type)", choices=["any","none","kmer","split","marker"])
    variety_list.add_argument("hasParents", type=bool, required=False, location="args", 
                              help="parent(s) known for this variety")
    variety_list.add_argument("hasOffspring", type=bool, required=False, location="args", 
                              help="offspring known for this variety")

    variety_set = namespace.model("uid list to get varieties", {"uids": fields.List(fields.String, attribute="items", 
                              required=True, description="list of uids")})

    @namespace.doc(description="Get varieties")    
    @namespace.expect(variety_list)
    def get(self):
        try:
            start = int(request.args.get("start",0))
            number = int(request.args.get("number",10))        
            name = request.args.get("name",None)
            nameContains = request.args.get("nameContains",None)
            origin = request.args.get("origin",None)
            year = request.args.get("year",None)
            collection = request.args.get("collection",None)
            dataType = request.args.get("dataType",None)
            hasParents = request.args.get("hasParents",None)
            hasOffspring = request.args.get("hasOffspring",None)
            condition_sql = "1"
            condition_variables = []
            if not name==None:
                condition_sql = condition_sql + " AND ((`variety`.`name` LIKE ?) OR (`condition_synonym`.`synonym` LIKE ?))"
                condition_variables.extend([name,name])
            if not origin==None:
                origin_list = origin.split(",")
                condition_sql = condition_sql + " AND (`variety`.`origin` IN ("+",".join(["?"]*len(origin_list))+"))"
                condition_variables.extend(origin_list)
            if not year==None:
                condition_sql = condition_sql + " AND (NOT (`variety`.`year_max` IS NULL AND `variety`.`year_min` IS NULL))"
                if yearPattern.match(year):
                    condition_sql = condition_sql + " AND ((`variety`.`year_max` >= ?) OR (`variety`.`year_max` IS NULL))"
                    condition_sql = condition_sql + " AND ((`variety`.`year_min` <= ?) OR (`variety`.`year_min` IS NULL))"
                    condition_variables.extend([year,year])
                elif yearMaxPattern.match(year):
                    condition_sql = condition_sql + " AND (`variety`.`year_max` < ?)"
                    condition_variables.append(year[1:5])
                elif yearMinPattern.match(year):
                    condition_sql = condition_sql + " AND (`variety`.`year_min` > ?)"
                    condition_variables.append(year[1:5])
                elif yearRangePattern.match(year):
                    year_min = year[0:4]
                    year_max = year[5:9]
                    condition_sql = condition_sql + " AND ((`variety`.`year_min` >= ?) OR "
                    condition_sql = condition_sql + "((`variety`.`year_min` IS NULL) AND (`variety`.`year_max` >= ?)))"
                    condition_sql = condition_sql + " AND ((`variety`.`year_max` <= ?) OR "
                    condition_sql = condition_sql + "((`variety`.`year_max` IS NULL) AND (`variety`.`year_min` <= ?)))"
                    condition_variables.extend([year_min,year_min,year_max,year_max])
                else:
                    abort(422, "incorrect year condition "+str(year))
            if not collection==None:
                collection_list = collection.split(",")
                condition_sql = condition_sql + " AND (`collection`.`uid` IN ("+",".join(["?"]*len(collection_list))+"))"
                condition_variables.extend(collection_list)
            if not dataType==None:
                if dataType=="any":
                    condition_sql = condition_sql + " AND NOT (`dataset`.`id` IS NULL)"  
                elif dataType=="none":
                    condition_sql = condition_sql + " AND (`dataset`.`id` IS NULL)"  
                elif dataType=="kmer":
                    condition_sql = condition_sql + " AND ((`dataset`.`type` IS 'kmer') OR (`dataset`.`type` IS 'split'))"  
                elif dataType=="split":
                    condition_sql = condition_sql + " AND (`dataset`.`type` IS 'split')" 
                elif dataType=="marker":
                    condition_sql = condition_sql + " AND (`dataset`.`type` IS 'marker')"
                else:
                    abort(422, "incorrect dataType condition "+str(dataset))
            if not hasParents==None:
                if _make_bool(hasParents):
                    condition_sql = condition_sql + " AND NOT (`ancestor`.`id` IS NULL)"
                else:
                    condition_sql = condition_sql + " AND (`ancestor`.`id` IS NULL)"       
            if not hasOffspring==None:
                if _make_bool(hasOffspring):
                    condition_sql = condition_sql + " AND NOT (`offspring`.`id` IS NULL)"
                else:
                    condition_sql = condition_sql + " AND (`offspring`.`id` IS NULL)"     
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("""SELECT COUNT(DISTINCT `variety`.`id`) AS `number` 
                            FROM `variety` 
                            LEFT JOIN `dataset` ON `variety`.`uid` = `dataset`.`variety` 
                            AND NOT `dataset`.`uid` IS NULL AND NOT `dataset`.`type` IS NULL 
                            LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` 
                            LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` 
                            LEFT JOIN `variety_ancestor` AS `ancestor` ON `variety`.`uid` = `ancestor`.`variety` 
                            LEFT JOIN `variety_ancestor` AS `offspring` ON `variety`.`uid` = `offspring`.`ancestor` 
                            LEFT JOIN `variety_synonym` AS `synonym` ON `variety`.`uid` = `synonym`.`uid` 
                            LEFT JOIN `variety_synonym` AS `condition_synonym` ON `variety`.`uid` = 
                                        `condition_synonym`.`uid` 
                            WHERE """ + condition_sql, tuple(condition_variables))  
            total = cursor.fetchone()[0]
            if start<total:
                cursor.execute("""SELECT `variety`.`uid`, `variety`.`type`, `variety`.`name`, `variety`.`origin`, 
                                `country`.`name` AS `country`, 
                                NULL AS `year`, `variety`.`year_min`, `variety`.`year_max`, 
                                GROUP_CONCAT(DISTINCT `synonym`.`synonym`) AS `synonyms`, 
                                `variety`.`species_id` AS `species_id`,
                                TRIM(`species`.`genus` || ' ' || `species`.`species`) AS `species_name`,
                                `species`.`genus` AS `species_genus`,
                                `species`.`species` AS `species_species`,
                                `variety`.`breeder_id` AS `breeder_id`,
                                `breeder`.`name` AS `breeder_name`,
                                COUNT(DISTINCT `dataset`.`id`) as `datasets` 
                                FROM `variety` 
                                LEFT JOIN `dataset` ON `variety`.`uid` = `dataset`.`variety` 
                                AND NOT `dataset`.`uid` IS NULL AND NOT `dataset`.`type` IS NULL
                                LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` 
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` 
                                LEFT JOIN `species` ON `variety`.`species_id` = `species`.`id` 
                                LEFT JOIN `breeder` ON `variety`.`breeder_id` = `breeder`.`id` 
                                LEFT JOIN `variety_ancestor` AS `ancestor` ON `variety`.`uid` = `ancestor`.`variety` 
                                LEFT JOIN `variety_ancestor` AS `offspring` ON `variety`.`uid` = `offspring`.`ancestor` 
                                LEFT JOIN `variety_synonym` AS `synonym` ON `variety`.`uid` = `synonym`.`uid` 
                                LEFT JOIN `variety_synonym` AS `condition_synonym` ON `variety`.`uid` = 
                                            `condition_synonym`.`uid` 
                                WHERE """+condition_sql+""" 
                                GROUP BY `variety`.`id` 
                                ORDER BY CASE `variety`.`type` WHEN 'species' THEN 0 ELSE 1 END, 
                                    `variety`.`name`, `variety`.`uid` 
                                LIMIT ?,?""",tuple(condition_variables + [start,number]))  
                resultList = [dict(row) for row in cursor.fetchall()]
            else:
                resultList = []
            for i in range(len(resultList)):
                resultList[i] = adjust_variety_response(resultList[i], collection, dataType, db_connection)                
            response = {"start": start, "number": number, "total": total, "list": resultList}
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
    @namespace.doc(description="Get multiple varieties by uid")    
    @namespace.expect(variety_set)
    def post(self):
        uids = namespace.payload.get("uids",[])
        try:
            if len(uids)>0:
                condition_sql = "`variety`.`uid` IN ("+(",".join(["?"]*len(uids)))+")"
                condition_variables = [] + uids;
                db_connection = haplotyping.service.API.get_db_connection()
                db_connection.row_factory = sqlite3.Row
                cursor = db_connection.cursor()
                cursor.execute("""SELECT COUNT(DISTINCT `variety`.`id`) AS `number` 
                                FROM `variety` 
                                LEFT JOIN `dataset` ON `variety`.`uid` = `dataset`.`variety` 
                                AND NOT (`dataset`.`uid` IS NULL) AND NOT (`dataset`.`type` IS NULL) 
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` 
                                LEFT JOIN `variety_synonym` AS `synonym` ON `variety`.`uid` = `synonym`.`uid` 
                                WHERE """+condition_sql, tuple(condition_variables))  
                total = cursor.fetchone()[0]
                cursor.execute("""SELECT `variety`.`uid`, `variety`.`name`, `variety`.`origin`, 
                                `country`.`name` AS `country`, 
                                NULL AS `year`, `variety`.`year_min`, `variety`.`year_max`, 
                                GROUP_CONCAT(DISTINCT `synonym`.`synonym`) AS `synonyms`, 
                                COUNT(DISTINCT `dataset`.`id`) as `datasets` 
                                FROM `variety` 
                                LEFT JOIN `dataset` ON `variety`.`uid` = `dataset`.`variety` 
                                AND NOT (`dataset`.`uid` IS NULL) AND NOT (`dataset`.`type` IS NULL) 
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` 
                                LEFT JOIN `variety_synonym` AS `synonym` ON `variety`.`uid` = `synonym`.`uid` 
                                WHERE """+condition_sql+""" 
                                GROUP BY `variety`.`id` 
                                ORDER BY `variety`.`name`, `variety`.`uid`""",tuple(condition_variables))  
                resultList = [dict(row) for row in cursor.fetchall()]
            else:
                total = 0
                resultList = []
            for i in range(len(resultList)):
                resultList[i] = adjust_variety_response(resultList[i], None, None, db_connection)                
            response = {"total": total, "list": resultList}
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>")
@namespace.doc(params={"uid": "unique identifier variety"})
class VarietyId(Resource):
    
    @namespace.doc(description="Get variety by uid")
    def get(self,uid):
        try:
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("""SELECT `variety`.`uid`, `variety`.`type`, `variety`.`name`, `variety`.`origin`, 
                            `country`.`name` AS `country`, 
                            GROUP_CONCAT(DISTINCT `synonym`.`synonym`) AS `synonyms`, 
                            NULL AS `year`, `variety`.`year_min`, `variety`.`year_max`, 
                            `variety`.`species_id` AS `species_id`,
                            TRIM(`species`.`genus` || ' ' || `species`.`species`) AS `species_name`,
                            `species`.`genus` AS `species_genus`,
                            `species`.`species` AS `species_species`,
                            `variety`.`breeder_id` AS `breeder_id`,
                            `breeder`.`name` AS `breeder_name`,
                            NULL as `datasets` 
                            FROM `variety` 
                            LEFT JOIN `dataset` ON `variety`.`uid` = `dataset`.`variety` 
                            AND NOT (`dataset`.`uid` IS NULL) AND NOT (`dataset`.`type` IS NULL) 
                            LEFT JOIN `species` ON `variety`.`species_id` = `species`.`id` 
                            LEFT JOIN `breeder` ON `variety`.`breeder_id` = `breeder`.`id` 
                            LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` 
                            LEFT JOIN `variety_synonym` AS `synonym` ON `variety`.`uid` = `synonym`.`uid` 
                            WHERE `variety`.`uid` = ? 
                            GROUP BY `variety`.`id`""",(uid,))  
            data = cursor.fetchone()
            if data:
                response = adjust_variety_response(dict(data), None, None, db_connection)
                return Response(json.dumps(response), mimetype="application/json")
            else:
                abort(404, "no variety with uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))


#!python

import os,logging,tempfile
from frictionless import Package
import gzip,csv,pandas as pd, numpy as np
import sqlite3, json, time
from sqlite3 import Error
import h5py, random, shutil

class ConstructDatabase:
    
    dbFilename = "db.sqlite"
    identifierBackup = "identifiers.json"
    markerDirectory = "marker"
    kmerDirectory = "kmer"
    
    def __init__(self, basedir:str, exportdir:str):
        #data
        self._baseDir = basedir
        self._exportDir = exportdir        
        #logging
        self._logger = logging.getLogger(__name__)
        #connection
        self._connection = None
        self._cursor = None
        self._clean()
        try:
            self._load_identifiers()
            self._connect_sqlite3()
            self._exportIndex()
            self._exportResources()
        except Error as e:
            self._logger.error(e)
        finally:
            self._close_sqlite3()
            self._store_identifiers()
            
    def _load_identifiers(self):
        if os.path.exists(os.path.join(self._exportDir,self.identifierBackup)):
            with open(os.path.join(self._exportDir,self.identifierBackup), mode="r") as f:
                self._identifiers = json.load(f)
                for resource in self._identifiers["collections"]["marker"].keys():
                    for internalId in self._identifiers["collections"]["marker"][resource].keys():
                        self._identifiers["collections"]["marker"][resource][internalId]["active"] = 0
                for resource in self._identifiers["collections"]["kmer"].keys():
                    for internalId in self._identifiers["collections"]["kmer"][resource].keys():
                        self._identifiers["collections"]["kmer"][resource][internalId]["active"] = 0
                for uid in self._identifiers["collections"]["uids"].keys():
                    for id in self._identifiers["collections"]["uids"][uid].keys():
                        self._identifiers["collections"]["uids"][uid][id]["active"] = 0
        else:
            self._identifiers = {"collections": {"uids": {}, "marker": {}, "kmer": {}}, "datasets": []}
            
    def _store_identifiers(self):
        with open(os.path.join(self._exportDir,self.identifierBackup), mode="w") as f:
            json.dump(self._identifiers,f)

    def _connect_sqlite3(self):
        self._logger.info("create {}".format(self.dbFilename))
        self._connection = sqlite3.connect(os.path.join(self._exportDir,self.dbFilename))  
        self._cursor = self._connection.cursor()
        
    def _close_sqlite3(self):
        if self._cursor:
            self._cursor.close()
        if self._connection:            
            self._connection.close()            
            
    def _clean(self):
        if os.path.exists(os.path.join(self._exportDir,self.markerDirectory)):
            self._logger.info("remove previous version {}".format(self.markerDirectory))
            shutil.rmtree(os.path.join(self._exportDir,self.markerDirectory))
        os.mkdir(os.path.join(self._exportDir,self.markerDirectory))
        if not os.path.exists(os.path.join(self._exportDir,self.kmerDirectory)):
            os.mkdir(os.path.join(self._exportDir,self.kmerDirectory))
        if os.path.exists(os.path.join(self._exportDir,self.dbFilename)):
            self._logger.info("remove previous version {}".format(self.dbFilename))
            os.remove(os.path.join(self._exportDir,self.dbFilename))
            
                
    def _exportIndex(self):  
        #add new
        indexFound = False
        for filename in sorted(os.listdir(self._baseDir)):
            if os.path.isfile(os.path.join(self._baseDir,filename)):
                if filename=="index.xlsx":
                    indexFound = True
                elif filename=="index.package.json":
                    package = Package(os.path.join(self._baseDir,filename))
                    #create structure
                    self._logger.info("initialise database")
                    sql = """CREATE TABLE IF NOT EXISTS "variety" (
                                "id" INTEGER NOT NULL,
                                "type" VARCHAR(7) CHECK("type" IN ("variety","species")) NOT NULL,
                                "uid" VARCHAR(13) NULL,
                                "name" VARCHAR(255) NULL,
                                "year_min" INTEGER NULL,
                                "year_max" INTEGER NULL,
                                "origin" VARCHAR(4) NULL,
                                "species_id" INTEGER NULL,
                                "breeder_id" INTEGER NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE INDEX "variety_name" ON "variety" ("name");
                            CREATE INDEX "variety_origin" ON "variety" ("origin");
                            CREATE INDEX "variety_species" ON "variety" ("species_id");
                            CREATE INDEX "variety_breeder" ON "variety" ("breeder_id");
                            CREATE UNIQUE INDEX "variety_uid" ON "variety" ("uid");
                            CREATE TABLE IF NOT EXISTS "variety_synonym" (
                                "id" INTEGER NOT NULL,
                                "uid" VARCHAR(13) NULL,
                                "synonym" VARCHAR(255) NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE INDEX "synonym_uid" ON "variety_synonym" ("uid");
                            CREATE TABLE IF NOT EXISTS "variety_ancestor" (
                                "id" INTEGER NOT NULL,
                                "variety" VARCHAR(13) NOT NULL,
                                "ancestor" VARCHAR(13) NULL,
                                "type" TEXT NULL,
                                "offspring" INTEGER NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE INDEX "variety_ancestor_ancestor" ON "variety_ancestor" ("ancestor");
                            CREATE INDEX "variety_ancestor_offspring" ON "variety_ancestor" ("offspring");
                            CREATE INDEX "variety_ancestor_variety" ON "variety_ancestor" ("variety");
                            CREATE TABLE IF NOT EXISTS "country" (
                                "id" INTEGER NOT NULL,
                                "uid" VARCHAR(4) NOT NULL,
                                "name" VARCHAR(255) NOT NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE UNIQUE INDEX "country_uid" ON "country" ("uid");
                            CREATE TABLE IF NOT EXISTS "species" (
                                "id" INTEGER NOT NULL,
                                "genus" VARCHAR(255) NOT NULL,
                                "species" VARCHAR(255) NOT NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE TABLE IF NOT EXISTS "breeder" (
                                "id" INTEGER NOT NULL,
                                "name" VARCHAR(255) NOT NULL,
                                "country" VARCHAR(4) NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE INDEX "breeder_country" ON "breeder" ("country");
                            CREATE TABLE IF NOT EXISTS "collection" (
                                "id" INTEGER NOT NULL,
                                "uid" VARCHAR(13) NULL,
                                "resource" VARCHAR(255) NULL,
                                "internal_id" INTEGER NOT NULL,
                                "type" VARCHAR(6) CHECK("type" IN ("marker","wgs","baits") ) NOT NULL,
                                "name" VARCHAR(255) NULL,
                                "experiment" VARCHAR(255) NULL,
                                "location" VARCHAR(255) NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE UNIQUE INDEX "collection_resource_internal_type" 
                                ON "collection" ("resource","internal_id","type");
                            CREATE UNIQUE INDEX "collection_uid" ON "collection" ("uid");
                            CREATE TABLE IF NOT EXISTS "dataset" (
                                "id" INTEGER NOT NULL,
                                "uid" VARCHAR(13) NULL,
                                "variety" VARCHAR(13) NULL,
                                "accession" VARCHAR(255) NULL,
                                "collection_id" INTEGER NOT NULL,
                                "internal_id" INTEGER NOT NULL,
                                "location" VARCHAR(255) NULL,
                                "type" VARCHAR(6) CHECK("type" IN ("marker","kmer","split") ) NULL,
                                "subtype" VARCHAR(6) CHECK("subtype" IN ("split","direct","read") ) NULL,
                                "version" VARCHAR(8) NULL,
                                PRIMARY KEY ("id")
                            );
                            CREATE INDEX "dataset_collection_id" ON "dataset" ("collection_id");
                            CREATE UNIQUE INDEX "dataset_collection_internal" 
                                ON "dataset" ("collection_id","internal_id");
                            CREATE UNIQUE INDEX "dataset_uid" ON "dataset" ("uid");
                            CREATE INDEX "dataset_variety" ON "dataset" ("variety");"""
                    self._connection.executescript(sql)
                    #countries
                    countries = pd.DataFrame(package.get_resource("index_countries").read_rows())
                    countries = countries[["code","name"]]
                    countries.columns = ["uid","name"]
                    countries.to_sql("country", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} countries".format(len(countries)))
                    #species
                    species = pd.DataFrame(package.get_resource("index_species").read_rows())
                    species = species[["id","genus","species"]]
                    species.to_sql("species", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} species".format(len(species)))
                    #breeders
                    breeders = pd.DataFrame(package.get_resource("index_breeders").read_rows())
                    breeders = breeders[["id","name","country"]]
                    breeders.to_sql("breeder", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} breeders".format(len(breeders)))
                    #varieties
                    varieties = pd.DataFrame(package.get_resource("index_varieties").read_rows())
                    varieties = varieties[["uid", "type","name", "yearMin", "yearMax", "origin", "speciesId","breederId"]]
                    varieties.columns = ["uid", "type", "name", "year_min", "year_max", "origin", "species_id", "breeder_id"]
                    varieties.to_sql("variety", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} varieties".format(len(varieties)))
                    #ancestors
                    ancestors = pd.DataFrame(package.get_resource("index_ancestors").read_rows())
                    ancestors = ancestors[["id","variety", "ancestor", "type", "offspring"]]
                    ancestors.to_sql("variety_ancestor", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} ancestors".format(len(ancestors)))
                    #synonyms
                    synonyms = pd.DataFrame(package.get_resource("index_synonyms").read_rows())
                    synonyms = synonyms[["uid","synonym"]]
                    synonyms.to_sql("variety_synonym", con=self._connection, index=False, if_exists="append")
                    self._logger.info("add {} synonyms".format(len(synonyms)))
                    return
        if indexFound:
            self._logger.error("no index package found, probably not validated")
        else:
            self._logger.error("no index found in {}".format(self._baseDir))
                    
                    
    def _exportResources(self):  
        #add new
        for filename in sorted(os.listdir(self._baseDir)):
            if os.path.isfile(os.path.join(self._baseDir,filename)):
                if filename=="index.package.json":
                    pass
                elif filename.endswith(".package.json"):                    
                    self._exportResource(os.path.join(self._baseDir,filename),os.path.basename(filename)[:-13])
                    self._logger.info("processed package {}".format(filename))
                    
    def _newCollectionUid(self, collectionType, resource, internalId):
        if collectionType in self._identifiers["collections"].keys():
            if resource in self._identifiers["collections"][collectionType]:
                if str(internalId) in self._identifiers["collections"][collectionType][resource]:
                    self._identifiers["collections"][collectionType][resource][str(internalId)]["active"] = 1
                    self._identifiers["collections"][collectionType][resource][str(internalId)]["used"] = time.time()
                    return self._identifiers["collections"][collectionType][resource][str(internalId)]["uid"]
        while True:
            uid = "DC_{}".format(str(random.randint(0,999999)).zfill(6))
            if not collectionType in self._identifiers["collections"].keys():
                self._identifiers["collections"][collectionType] = {}
            if not resource in self._identifiers["collections"][collectionType].keys():
                self._identifiers["collections"][collectionType][resource] = {}            
            if not uid in self._identifiers["collections"]["uids"]:
                now = time.time()
                self._identifiers["collections"][collectionType][resource][str(internalId)] = {
                    "active": 1, "used": now, "created": now, "uid": uid
                }
                self._identifiers["collections"]["uids"][uid] = {}
                return uid             
                        
    def _newDatasetUid(self, collectionUid, internalId):
        if collectionUid in self._identifiers["collections"]["uids"].keys():
            if str(internalId) in self._identifiers["collections"]["uids"][collectionUid]:
                self._identifiers["collections"]["uids"][collectionUid][str(internalId)]["active"] = 1
                self._identifiers["collections"]["uids"][collectionUid][str(internalId)]["used"] = time.time()
                return self._identifiers["collections"]["uids"][collectionUid][str(internalId)]["uid"]
            else:
                while True:
                    uid = "DS_{}".format(str(random.randint(0,999999)).zfill(6))
                    if not uid in self._identifiers["datasets"]:
                        now = time.time()
                        self._identifiers["collections"]["uids"][collectionUid][str(internalId)] = {
                            "active": 1, "used": now, "created": now, "uid": uid
                        }
                        self._identifiers["datasets"].append(uid)
                        return uid 
        else:
            raise Exception("unknown collection uid {}".format(collectionUid))
                        
    def _createCollection(self, resource, internal_id, collection_type, resource_name, experiment_name, location):
        uid = self._newCollectionUid(collection_type, resource, internal_id)
        query = """INSERT OR IGNORE INTO `collection` (`uid`,`resource`,`internal_id`,`type`,`name`,`experiment`,`location`) 
        VALUES (?,?,?,?,?,?,?)"""
        self._connection.execute(query, (uid,resource, int(internal_id), collection_type, 
                                         resource_name, experiment_name, location))
        self._connection.commit()
        query = """SELECT `id` FROM `collection` WHERE `resource` = ? AND `internal_id` = ? AND `type` = ?"""
        self._cursor.execute(query, (resource, int(internal_id),collection_type,))  
        collection = self._cursor.fetchone()
        if collection:
            return (collection[0],uid,)
        else:
            raise Exception("could not create collection")
        
    def _createDataset(self,variety,accession,collection_id,collection_uid,internal_id,
                       location,type=None,subtype=None,version=None):
        uid = self._newDatasetUid(collection_uid, internal_id)
        query = """INSERT OR IGNORE INTO `dataset` (`uid`,`variety`,`accession`,`collection_id`,`internal_id`,
                            `location`,`type`,`subtype`,`version`) 
        VALUES (?,?,?,?,?,?,?,?,?)"""
        self._connection.execute(query, (uid,variety, accession, int(collection_id), int(internal_id), 
                                         location, type, subtype, version))
        self._connection.commit()
        
    def _getUint(maximumValue):
        if maximumValue<=np.iinfo(np.uint8).max:
            return "uint8"
        elif maximumValue<=np.iinfo(np.uint16).max:
            return "uint16"
        elif maximumValue<=np.iinfo(np.uint32).max:
            return "uint32"
        else:
            return "uint64"
    
    def _getMaxLengthColumn(df,name):
        return max(1,max([len(item) if item else 0 for item in df[name]]))
    
    def _exportResource(self, fileName:str, resourceName:str):
        package = Package(fileName)        
        if package.has_resource("varieties") and package.has_resource("metadata") and package.has_resource("experiments"):
            if (package.has_resource("sequences") or 
                    (package.has_resource("scores") 
                     and package.has_resource("markers") 
                     and package.has_resource("mappings"))):
                #get shared data
                metadata = pd.DataFrame(package.get_resource("metadata").read_rows())
                metadata = {x:y for (x,y,) in zip(metadata["label"],metadata["value"])}
                varieties = pd.DataFrame(package.get_resource("varieties").read_rows())
                experiments = pd.DataFrame(package.get_resource("experiments").read_rows())
                if len(varieties)>0 and len(experiments)>0:
                    varieties = varieties.set_index("id")
                    experiments = experiments.set_index("id")
                    self._logger.info("store data for '{}'".format(metadata.get("name",resourceName)))
                else:
                    return
                #process sequence data
                if package.has_resource("sequences"):
                    collectionList = {}
                    sequences = pd.DataFrame(package.get_resource("sequences").read_rows())
                    if len(sequences)>0:
                        for experiment_id in set(sequences["experiment_id"]):
                            collectionId=None
                            for id,row in sequences[sequences["experiment_id"]==experiment_id].iterrows():
                                variety = varieties.loc[row["variety_id"]]
                                if variety["uid"]:
                                    #use seperate collection for each sequence type and id
                                    if ((len(collectionList)==0) or 
                                        (not row["type"] in collectionList) or
                                        ((row["type"] in collectionList) and 
                                         (not experiment_id in collectionList[row["type"]]))
                                       ):
                                        experiment_name = experiments.loc[experiment_id]["name"]
                                        locationResource = metadata.get("location","")
                                        locationResource = "" if locationResource==None else locationResource
                                        locationExperiment = experiments.loc[experiment_id]["location"]
                                        locationExperiment = "" if locationExperiment==None else locationExperiment
                                        location = os.path.join(locationResource,locationExperiment)
                                        (collectionId,collectionUid) = self._createCollection(
                                             resourceName, experiment_id, row["type"],
                                             metadata.get("name",resourceName), experiment_name, location)
                                        if not row["type"] in collectionList:
                                            collectionList[row["type"]] = {}
                                        collectionList[row["type"]][experiment_id] = {"collectionId": collectionId, 
                                                                       "collectionUid": collectionUid}
                                    self._createDataset(variety["uid"],variety["accession"],
                                                        collectionList[row["type"]][experiment_id]["collectionId"],
                                                        collectionList[row["type"]][experiment_id]["collectionUid"],
                                                        row["variety_id"],row["location"])

                #process marker data
                if (package.has_resource("scores") and package.has_resource("markers") and package.has_resource("mappings")):
                    scores = pd.DataFrame(package.get_resource("scores").read_rows())
                    markers = pd.DataFrame(package.get_resource("markers").read_rows())
                    if len(markers)>0:
                        markers = markers.set_index("id")
                    mappings = pd.DataFrame(package.get_resource("mappings").read_rows())
                    #process scores
                    for score_id,row in scores.iterrows():
                        varietyList = []
                        markerList = []
                        resourceData = {"data":[],"lg":[],"marker":[],"variety":[], "varietyIndex": {}}
                        with tempfile.NamedTemporaryFile() as tmpMarkerFile:
                            sourceFilename = os.path.join(package.basepath,row["source"])
                            open_fn = gzip.open if sourceFilename.endswith(".gz") else open
                            #get all data from external file
                            with open_fn(sourceFilename, "rt") as f: 
                                reader = csv.reader(f, delimiter=",")
                                line = list(next(reader))
                                markerList = [int(item) for item in line[1:]]
                                for line in reader:
                                    items = list(line)
                                    variety_id = int(items[0])
                                    resourceData["data"].append([int(item) if item.isnumeric() else -1 for item in line[1:]])
                                    varietyList.append(variety_id)
                            if len(varietyList)>0 and len(markerList)>0:
                                experiment_id = row["experiment_id"]
                                experiment_location = experiments.loc[experiment_id]["location"]
                                experiment_location = "" if experiment_location==None else experiment_location
                                markerFilename = "markers_{}_{}_{}.h5".format(resourceName,experiment_id,score_id)
                                os.makedirs(os.path.join(self._exportDir,self.markerDirectory,
                                                         experiment_location), exist_ok=True)
                                self._logger.info("create marker database '{}'".format(markerFilename))
                                hf = h5py.File(os.path.join(self._exportDir,self.markerDirectory,
                                                            experiment_location,markerFilename), "w")
                                #data
                                hf.create_dataset("data", data=resourceData["data"], dtype=h5py.h5t.NATIVE_INT8)
                                #store linkage groups (chromosomes) and make an index to get assigned identifier later
                                lgs = [item for item in set(mappings["lg"]) if item]
                                lgIndex = {}
                                resourceData["lg"].append(("",))
                                for item in sorted(lgs):
                                    lgIndex[item]=len(resourceData["lg"])
                                    resourceData["lg"].append((item,))                                    
                                dtypeLgsList=[
                                    ("name","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(mappings,"lg")))
                                 ]
                                dtLgs=np.dtype(dtypeLgsList)
                                dsLgs=hf.create_dataset("lg", data=resourceData["lg"], dtype=dtLgs, chunks=None)
                                #mapping
                                mappingList = list(set(mappings["mapping"]))
                                dtypeMappingsList=[
                                ]
                                for mapping in mappingList:
                                    dtypeMappingsList.append(
                                        (mapping,[("lg",ConstructDatabase._getUint(len(resourceData["lg"])),),
                                                  ("pos","uint32"),]))                                    
                                dtMappings=np.dtype(dtypeMappingsList)
                                dsMappings=hf.create_dataset("mapping",(len(markerList),), 
                                                      dtype=dtMappings, chunks=None)
                                for mapping in mappingList:
                                    mappingsSubset = mappings[mappings["mapping"]==mapping]
                                    mappingsSubset = mappingsSubset[mappingsSubset["marker_id"].isin(markerList)]
                                    if len(mappingsSubset)>0:
                                        mappingsSubset = mappingsSubset.set_index("marker_id")
                                        for mappingId,mappingRow in mappingsSubset.iterrows():
                                            lgId = lgIndex.get(mappingRow["lg"],0)
                                            if not np.isnan(mappingRow["pos"]):
                                                dsMappings[markerList.index(mappingId),mapping] = (
                                                    lgId,mappingRow["pos"],)
                                #markers
                                for id,row in markers.loc[markerList].iterrows():
                                    entry = (row["name"],row["ref"],row["alt"],
                                                   row["sequence_ref"],row["sequence_alt"],
                                                   row["sequence_left"],row["sequence_right"],)
                                    entry = tuple([item if item else "" for item in entry])
                                    resourceData["marker"].append(entry)
                                dtypeMarkersList=[
                                    ("name","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(markers,"name"))),
                                    ("ref","S1"),
                                    ("alt","S1"),
                                    ("sequenceRef","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(markers,"sequence_ref"))),
                                    ("sequenceAlt","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(markers,"sequence_alt"))),
                                    ("sequenceLeft","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(markers,"sequence_left"))),
                                    ("sequenceRight","S{}".format(
                                        ConstructDatabase._getMaxLengthColumn(markers,"sequence_right")))
                                 ]
                                dtMarkers=np.dtype(dtypeMarkersList)
                                dsMarkers=hf.create_dataset("marker",data=resourceData["marker"], 
                                                      dtype=dtMarkers, chunks=None)
                                #varieties
                                for id,row in varieties.loc[varietyList].iterrows():
                                    resourceData["varietyIndex"][id] = len(resourceData["variety"])
                                    resourceData["variety"].append((row["name"] if row["name"] else "",))
                                varietyNameLength = max(1,max([len(item) for item in varieties["name"]]))
                                dtypeVarietiesList=[
                                    ("name","S{}".format(varietyNameLength))
                                 ]
                                dtVarieties=np.dtype(dtypeVarietiesList)
                                dsVarieties=hf.create_dataset("variety",data=resourceData["variety"], 
                                                      dtype=dtVarieties, chunks=None)
                                hf.close()
                                self._logger.info("register entries database '{}'".format(markerFilename))
                                #register markers in sqlite database
                                experiment_name = experiments.loc[experiment_id]["name"]
                                experiment_location = experiments.loc[experiment_id]["location"]
                                experiment_location = "" if experiment_location==None else experiment_location
                                (collectionId,collectionUid) = self._createCollection(resourceName, experiment_id, "marker",
                                                     metadata.get("name",resourceName), 
                                                     experiment_name, experiment_location)
                                for id,row in varieties.loc[varietyList].iterrows():
                                    internal_id = resourceData["varietyIndex"][id]
                                    self._createDataset(row["uid"], row["accession"], collectionId, collectionUid,
                                                            internal_id, markerFilename, "marker")
                                    
                            
                            
                             
                    

            
                    
                    




#!python

import os,logging
import sqlite3
from sqlite3 import Error
import h5py

class CheckDatabase:
    
    dbFilename = "db.sqlite"
    markerDirectory = "marker"
    kmerDirectory = "kmer"
    
    def __init__(self, dbdir:str, markerdir:str = None, kmerdir:str = None):
        #data
        self._dbDir = dbdir
        self._markerDir = os.path.join(self._dbDir,self.markerDirectory) if markerdir==None else markerdir
        self._kmerDir = os.path.join(self._dbDir,self.kmerDirectory) if kmerdir==None else kmerdir
        #logging
        self._logger = logging.getLogger(__name__)
        #check directories
        if not os.path.exists(self._markerDir):
            self._logger.error("marker directory {} not found".format(self._markerDir))
            return
        if not os.path.exists(self._kmerDir):
            self._logger.error("kmer directory {} not found".format(self._kmerDir))
            return
        #connection
        self._connection = None
        self._cursor = None
        try:
            self._connect_sqlite3()
            self._reset_dataset_types()
            self._check_markers()
            self._check_kmers()
        except Error as e:
            print(e)
        finally:
            self._close_sqlite3()

    def _connect_sqlite3(self):
        if os.path.exists(os.path.join(self._dbDir,self.dbFilename)):
            self._logger.info("open connection {}".format(self.dbFilename))
            self._connection = sqlite3.connect(os.path.join(self._dbDir,self.dbFilename))  
            self._connection.row_factory = sqlite3.Row
            self._cursor = self._connection.cursor()
        else:
            raise Exception("no {} found".format(self.dbFilename))
        
    def _close_sqlite3(self):
        if self._cursor:
            self._cursor.close()
        if self._connection:            
            self._connection.close()
            
    def _reset_dataset_types(self):
        self._logger.debug("reset dataset types")
        query = """UPDATE `dataset` SET `type` = NULL"""
        self._connection.execute(query)
        self._connection.commit()
            
    def _check_markers(self):
        query = """SELECT * FROM `collection` WHERE `type` = 'marker' ORDER BY `id`"""
        self._cursor.execute(query)
        collections = self._cursor.fetchall()        
        for row in collections:
            collection = dict(zip(row.keys(), row))
            if collection["location"]!=None:
                if not os.path.exists(os.path.join(self._markerDir,collection["location"])):
                    self._logger.error("marker collection {}: location {} not found".format(
                        collection["name"], collection["location"]))
                    continue
                else:
                    self._logger.info("marker collection {}: check datasets".format(collection["name"]))
            #get datasets
            query = """SELECT * FROM `dataset` WHERE `collection_id` = ? ORDER BY `id`"""
            self._cursor.execute(query, (collection["id"],))  
            datasets = self._cursor.fetchall()
            markerDbs = {}
            for row in datasets:
                dataset = dict(zip(row.keys(), row))
                if dataset["location"]!=None:
                    if not dataset["location"] in markerDbs.keys():
                        if not collection["location"]==None:
                            dbLocation = os.path.join(self._markerDir,collection["location"],dataset["location"])
                        else:
                            dbLocation = os.path.join(self._markerDir,dataset["location"])
                        if not os.path.exists(dbLocation):   
                            self._logger.error("marker collection {}: database {} not found".format(
                                collection["name"], dbLocation))
                            markerDbs[dataset["location"]] = []
                        else:
                            with h5py.File(dbLocation, "r") as hf:
                                markerDbs[dataset["location"]] = list(range(len(hf["variety"])))
                    if dataset["internal_id"] in markerDbs[dataset["location"]]:
                        query = """UPDATE `dataset` SET `type` = 'marker' WHERE `id` = ? AND `collection_id` = ?"""
                        self._cursor.execute(query, (dataset["id"], collection["id"],))  
                    else:
                        self._logger.error("marker collection {}: markers for {} not found in database {}".format(
                            collection["name"], dataset["variety"], dataset["location"]))
            self._connection.commit()
            
    def _check_kmers(self):
        query = """SELECT * FROM `collection` WHERE `type` = 'wgs' OR `type` = 'baits' ORDER BY `id`"""
        self._cursor.execute(query)
        collections = self._cursor.fetchall()        
        for row in collections:
            collection = dict(zip(row.keys(), row))
            if collection["location"]!=None:
                if not os.path.exists(os.path.join(self._kmerDir,collection["location"])):
                    self._logger.error("kmer collection {}: location {} not found".format(
                        collection["name"], collection["location"]))
                    continue
                else:
                    self._logger.info("kmer collection {}: check datasets".format(collection["name"]))
            #get datasets
            query = """SELECT * FROM `dataset` WHERE `collection_id` = ? ORDER BY `id`"""
            self._cursor.execute(query, (collection["id"],))  
            datasets = self._cursor.fetchall()
            locationsNotFound = 0
            kmerNotFound = 0
            splitNotFound = 0
            for row in datasets:
                dataset = dict(zip(row.keys(), row))
                if dataset["location"]!=None:
                    if not collection["location"]==None:
                        kmerLocation = os.path.join(self._kmerDir,
                                                    collection["location"],
                                                    dataset["location"])
                    else:
                        kmerLocation = os.path.join(self._kmerDir,
                                                    dataset["location"])
                    if not os.path.exists(kmerLocation):   
                        self._logger.debug("kmer collection {}: location {} not found".format(
                                collection["name"], kmerLocation))
                        locationsNotFound+=1
                    else:
                        if not os.path.exists(os.path.join(kmerLocation,"kmer.kmc.kmc_pre")):  
                            kmerNotFound+=1
                        elif not os.path.exists(os.path.join(kmerLocation,"kmer.kmc.kmc_suf")):  
                            kmerNotFound+=1
                        else:
                            datasetType = "kmer"
                            datasetSubType = None
                            datasetVersion = None
                            dbLocation = os.path.join(kmerLocation,"kmer.data.h5")
                            if not os.path.exists(dbLocation): 
                                splitNotFound+=1
                            else:
                                try:
                                    with h5py.File(dbLocation, "r") as h5file:                                        
                                        datasetVersion = h5file["/config"].attrs["version"]
                                        if not ("/relations" in h5file and "/relations/direct" in h5file):
                                            datasetSubType = "split"
                                        elif not ("/relations/readData" in h5file):
                                            datasetSubType = "direct"
                                        else:
                                            datasetSubType = "read"
                                        datasetType = "split"
                                except:
                                    self._logger.error("couldn't parse {}".format(dbLocation))
                            query = """UPDATE `dataset` SET `type` = ?, `subtype` = ?, `version` = ?
                                        WHERE `id` = ? AND `collection_id` = ?"""
                            self._cursor.execute(query, (datasetType, datasetSubType, datasetVersion, 
                                                         dataset["id"], collection["id"],))  
            self._connection.commit()            
            if locationsNotFound>0:
                self._logger.error("kmer collection {}: {} location(s) not found".format(
                                collection["name"], locationsNotFound))
            if kmerNotFound>0:
                self._logger.error("kmer collection {}: {} k-mer databases not found".format(
                                collection["name"], kmerNotFound))
            if splitNotFound>0:
                self._logger.error("kmer collection {}: {} splitting k-mer databases not found".format(
                                collection["name"], splitNotFound))
                            
                    

                
            
            
    




import logging, h5py, os, sys, glob, re, math, shutil
import numpy as np, tables, gzip, csv
import multiprocessing as mp

import haplotyping
import haplotyping.index.splits
import haplotyping.index.connections

class Database:
    
    """
    Constructing splitting k-mer database
    
    Parameters
    ----------------------
    k : int
        The used k-mer size, should correspond with the k-mer size of the sorted k-mer list

    name : str
        Name of the variety

    filenameBase : str
        Base name and location for the splitting k-mer database and the temporary databases

    sortedIndexFile: str
        Location of the sorted k-mer list
        The k-mer size for k-mers in this list should correspond with the provided k parameter

    readFiles: optional, default is empty list
        A list with locations from readFiles
        Optional, but at least readFiles or pairedReadFiles should be defined and non-empty

    pairedReadFiles: optional, default is empty list
        A list with pairs of locations from paired readFiles
        Optional, but at least readFiles or pairedReadFiles should be defined and non-empty

    minimumFrequency: int, optional, default is 2
        The minimum frequency for a splitting k-mer.
        Adjusting this number will change the amount of read error based results but 
        can also introduce gaps; With higher read depth or lower error rate, this number can possibly be increased
        
    maximumMemory: int, default is 0 (for no maximum)
        Maximum memory size, reduces the number of processes if too much memory is needed
        
    maximumProcesses: int, default is 0 (for no maximum)
        The maximum number of processes including main process, should be at least 4
        
    automatonKmerSize: int, optional, default is 0 (for automatically)
        The used reduced k-mer size, maximum is the k-mer size
        
    indexType: str, optional, default is None (for a full index)
        Possible values:
        - "onlySplittingKmers": Only the first step will be executed, using a single process.
        - "onlyDirectConnections": Only direct connections are extracted from the reads
        
    debug: bool, optional, default is False
        Only use this when debugging or extending the code.      
        
    keepTemporaryFiles: bool, optional, default is False
        Only use this when debugging or extending the code.      
        
    """
    
    #define index types
    FULLINDEX = "full"
    ONLYSPLITTINGKMERS = "onlySplittingKmers"
    ONLYDIRECTCONNECTIONS = "onlyDirectConnections"

    def __init__(self,
                 k: int, 
                 name: str, 
                 filenameBase: str,
                 sortedIndexFile: str,
                 readFiles=[], pairedReadFiles=[],
                 minimumFrequency: int = 2,
                 maximumMemory: int = 0,
                 maximumProcesses: int = 0,
                 automatonKmerSize: int = 0,
                 indexType: str = None,
                 debug: bool = False,
                 keepTemporaryFiles: bool=False):  
        
        """
        Internal use only: initialize
        """
        
        #use spawn, also on unix
        if __name__ == "__main__":
            mp.set_start_method("spawn")
    
        if mp.get_start_method()=="spawn":
            frame = sys._getframe()
            while frame:
                if "__name__" in frame.f_locals.keys():
                    if not frame.f_locals["__name__"]=="__main__":
                        return
                    break                    
                frame = frame.f_back
        
        #logger
        self._logger = logging.getLogger(__name__)
        self._logger.info("create data storage for {} in {}".format(name,filenameBase))
        self._logger.debug("haplotyping package version {}".format(haplotyping._version.__version__))
                          
        
        #store variables
        self.k=k
        self.name=name
        self.debug = debug
        if indexType==self.ONLYSPLITTINGKMERS or indexType==self.ONLYDIRECTCONNECTIONS:
            self.indexType = indexType
        elif indexType=="" or indexType==None or indexType==self.FULLINDEX:
            self.indexType = self.FULLINDEX
        else:
            raise Exception("unknown indexType '{}'".format(indexType))
        self.version = haplotyping._version.__version__
        self.automatonKmerSize = automatonKmerSize
        self.minimumFrequency = minimumFrequency
        self.keepTemporaryFiles = keepTemporaryFiles
        self.filenameBase = filenameBase
        self.maximumMemory = maximumMemory
        self.maximumProcesses = maximumProcesses
                
        #check boundaries number of processes
        assert self.automatonKmerSize>=0 and self.automatonKmerSize<=self.k
        assert self.maximumMemory>=0
        assert self.maximumProcesses>=0
        
        if (not self.indexType == self.ONLYSPLITTINGKMERS) and (len(readFiles)==0) and (len(pairedReadFiles)==0):
            self._logger.error("no read files provided")
        else:                
            #define filenames
            filename = "{}.h5".format(filenameBase)
            
            #skip finished steps
            if self.debug:
                filename_splits = "{}.splits.h5".format(filenameBase)
                filename_distances = "{}.distances.h5".format(filenameBase)
                if os.path.exists(filename_distances):
                    shutil.copyfile(filename_distances, filename)
                elif os.path.exists(filename_splits):
                    shutil.copyfile(filename_splits, filename)

            #use tables for temporary file, and h5py for final
            #don't use libver="latest" before checking other tools!
            with h5py.File(filename,"a") as h5file:

                #set or check config
                if not "/config" in h5file:
                    h5file.create_group("/config")
                    h5file["/config"].attrs["k"] = self.k
                    h5file["/config"].attrs["version"] = self.version
                    h5file["/config"].attrs["indexType"] = self.indexType
                    h5file["/config"].attrs["automatonKmerSize"] = self.automatonKmerSize
                    h5file["/config"].attrs["minimumCanonicalSplitFrequency"] = self.minimumFrequency
                    h5file.flush()
                else:
                    assert h5file["/config"].attrs["k"] == self.k
                    assert h5file["/config"].attrs["version"] == self.version
                    assert h5file["/config"].attrs["indexType"] == self.indexType
                    assert h5file["/config"].attrs["automatonKmerSize"] == self.automatonKmerSize
                    assert h5file["/config"].attrs["minimumCanonicalSplitFrequency"] == self.minimumFrequency
                    
                #these settings are allowed to change in secondary runs
                h5file["/config"].attrs["name"] = self.name
                h5file["/config"].attrs["debug"] = self.debug
                h5file["/config"].attrs["maximumMemory"] = self.maximumMemory
                h5file["/config"].attrs["maximumProcesses"] = self.maximumProcesses
                
                #get splitting k-mers from index   
                if not ("/split" in h5file and "/histogram" in h5file):
                    if not os.path.exists(sortedIndexFile):
                        self._logger.error("no sorted k-mer list provided")
                    else:
                        self._logger.debug("get splitting k-mers from the provided index")
                        haplotyping.index.splits.Splits(sortedIndexFile, h5file, 
                                                        self.filenameBase, self.debug, self.keepTemporaryFiles)    
                        h5file.flush()
                        #backup
                        if self.debug:
                            shutil.copyfile(filename, filename_splits)
                else:
                    self._logger.debug("detected splitting k-mers from previous run")
                    
                #parse read files and store distances
                if (not self.indexType == self.ONLYSPLITTINGKMERS) and ("/split" in h5file) and ("/histogram" in h5file):
                    if not ("/relations" in h5file and "/connections" in h5file):
                        self._logger.debug("parse read files and store distances in database")
                        haplotyping.index.connections.Connections(readFiles,pairedReadFiles, h5file, 
                                                      self.filenameBase, self.indexType, 
                                                      self.debug, self.keepTemporaryFiles)
                        h5file.flush()
                        #backup
                        if self.debug:
                            shutil.copyfile(filename, filename_distances)
                    else:
                        self._logger.debug("detected distances from previous run")

                #finished
                h5file.flush()
                
                if self.debug and not self.keepTemporaryFiles:
                    if os.path.exists(filename_distances):
                        os.remove(filename_distances)
                    if os.path.exists(filename_splits):
                        os.remove(filename_splits)

                
    letters = ["A","C","G","T"]
        
    def detectReadFiles(location: str, recursive=True):
        unpairedReadFiles = []
        pairedReadFiles = []
        #get files
        patterns = ["*.fq.gz","*.fastq.gz"]
        if recursive:
            allReadFiles = [y for x in os.walk(location) 
                for p in patterns
                for y in glob.glob(os.path.join(x[0], p))]
        else:
            allReadFiles = [y for p in patterns
                for y in glob.glob(os.path.join(location, p))]
        allReadFiles.sort()
        #split between regular and paired
        processed = set()
        for filename in allReadFiles:
            if filename in processed:
                pass
            else:
                basename = os.path.basename(filename)
                dirname = filename[:len(filename)-len(basename)]  
                if re.search(r'_R1_001\.', basename):
                    filename0 = dirname + basename
                    filename1 = dirname + re.sub(r'_R1_001\.', 
                                                 '_R2_001.', basename)
                    if filename0==filename and filename1 in allReadFiles and not filename in processed:
                        processed.add(filename0)
                        processed.add(filename1)
                        pairedReadFiles.append((filename0,filename1,))
                    else:
                        processed.add(filename)
                        unpairedReadFiles.append(filename)
                elif re.search(r'_1P\.', basename):
                    filename0 = dirname + basename
                    filename1 = dirname + re.sub(r'_1P\.', 
                                                 '_2P.', basename)
                    if filename0==filename and filename1 in allReadFiles and not filename in processed:
                        processed.add(filename0)
                        processed.add(filename1)
                        pairedReadFiles.append((filename0,filename1,))
                    else:
                        processed.add(filename)
                        unpairedReadFiles.append(filename)
                elif re.search(r'_R1\.', basename):
                    filename0 = dirname + basename
                    filename1 = dirname + re.sub(r'_R1\.', 
                                                 '_R2.', basename)
                    if filename0==filename and filename1 in allReadFiles and not filename in processed:
                        processed.add(filename0)
                        processed.add(filename1)
                        pairedReadFiles.append((filename0,filename1,))
                    else:
                        processed.add(filename)
                        unpairedReadFiles.append(filename)
                else:
                    processed.add(filename)
                    unpairedReadFiles.append(filename)

        return (unpairedReadFiles, pairedReadFiles, allReadFiles)
    
    def detectKmerSize(location: str):
        try:
            with gzip.open(location, "rt") as f: 
                reader = csv.reader(f, delimiter="\t")
                for line in reader:
                    return len(str(line[0]))
        except OSError as ex:
            raise Exception("problem with sorted list: {}".format(ex))
    
    def getTablesUint(maximumValue, position):
        if maximumValue<=np.iinfo(np.uint8).max:
            return tables.UInt8Col(pos=position)
        elif maximumValue<=np.iinfo(np.uint16).max:
            return tables.UInt16Col(pos=position)
        elif maximumValue<=np.iinfo(np.uint32).max:
            return tables.UInt32Col(pos=position)
        else:
            return tables.UInt64Col(pos=position)
        
    def getTablesUintAtom(maximumValue):
        if maximumValue<=np.iinfo(np.uint8).max:
            return tables.UInt8Atom()
        elif maximumValue<=np.iinfo(np.uint16).max:
            return tables.UInt16Atom()
        elif maximumValue<=np.iinfo(np.uint32).max:
            return tables.UInt32Atom()
        else:
            return tables.UInt64Atom()
        
    def getUint(maximumValue):
        if maximumValue<=np.iinfo(np.uint8).max:
            return "uint8"
        elif maximumValue<=np.iinfo(np.uint16).max:
            return "uint16"
        elif maximumValue<=np.iinfo(np.uint32).max:
            return "uint32"
        else:
            return "uint64"




import h5py, haplotyping, numpy as np

class Split:
    
    #---
    
    def _findItem(item,table,start=0,number=None, cache={}):
        if number==None:
            number=table.shape[0]
        itemBinary = bytearray(item,"utf8")
        newCache={}
        minRowId = start
        maxRowId = number-1
        for key in cache.keys():
            if cache[key]>=minRowId:
                if key<itemBinary:
                    minRowId=max(minRowId,cache[key])
                elif key>itemBinary:
                    maxRowId=min(maxRowId,cache[key])
                    newCache[key]=cache[key]
                else:
                    minRowId = cache[key]
                    maxRowId = cache[key]
                    newCache[key]=cache[key]
        currentRowId = minRowId + int((maxRowId-minRowId)/2) 
        while True:
            if start>=number:
                return (None,currentRowId,newCache,)
            currentRow = table[currentRowId]
            currentItem = currentRow[0]        
            if currentItem==itemBinary:
                newCache[currentItem] = currentRowId
                return (currentRow,currentRowId,newCache,)
            elif currentItem>itemBinary:
                newCache[currentItem] = currentRowId
                maxRowId = currentRowId-1            
            else:
                minRowId = currentRowId+1
            if maxRowId<minRowId:
                return (None,currentRowId,newCache,)
            else:
                currentRowId = minRowId + int((maxRowId-minRowId)/2)
                
    #old code to get correct entries without index
    def _findId(id,table,start=0,number=None, cache={}):
        if number==None:
            number=table.shape[0]
        newCache={}
        minRowId = start
        maxRowId = number-1
        for cacheId in cache.keys():
            if cache[cacheId]>=minRowId:
                if cacheId<id:
                    minRowId=max(minRowId,cache[cacheId])
                elif cacheId>id:
                    maxRowId=min(maxRowId,cache[cacheId])
                    newCache[cacheId]=cache[cacheId]
                else:
                    minRowId = cache[cacheId]
                    maxRowId = cache[cacheId]
                    newCache[cacheId]=cache[cacheId]
        currentRowId = minRowId + int((maxRowId-minRowId)/2) 
        while True:
            if start>=number:
                return ([],currentRowId,newCache,)
            currentRow = table[currentRowId]
            currentId = currentRow[0][0]    
            if currentId==id:
                currentRows = [currentRow]
                tmpId = currentRowId
                while tmpId>0:
                    tmpId-=1
                    tmpRow = table[tmpId]
                    if tmpRow[0][0]==id:
                        currentRows.append(tmpRow)
                    else:
                        break
                tmpId = currentRowId
                while tmpId<table.shape[0]-1:
                    tmpId+=1
                    tmpRow = table[tmpId]
                    if tmpRow[0][0]==id:
                        currentRows.append(tmpRow)
                    else:
                        break
                newCache[currentId] = currentRowId
                return (currentRows,currentRowId,newCache,)
            elif currentId>id:
                newCache[currentId] = currentRowId
                maxRowId = currentRowId-1            
            else:
                minRowId = currentRowId+1
            if maxRowId<minRowId:
                return ([],currentRowId,newCache,)
            else:
                currentRowId = minRowId + int((maxRowId-minRowId)/2)
                
    #---
    
    def _translate_type(value):
        if value=="l":
            return "left"
        elif value=="r":
            return "right"
        elif value=="b":
            return "both"
        else:
            return value
    
    def _base_result(row, h5file: h5py.File, expandCkmer=True):
        if row:
            response = {
                "base": row[0].decode("ascii"),
                "number": int(row[1])
            }
            branches = row.dtype[2].names
            ckmerTable = h5file.get("/split/ckmer")
            if expandCkmer:
                response["ckmers"] = []
                for i in range(len(branches)):
                    if row[2][i][0]>0:
                        ckmerRow = ckmerTable[row[2][i][1]]
                        response["ckmers"].append(Split._kmer_result(ckmerRow,h5file,False))
            else:                
                response["ckmers"] = {}
                for i in range(len(branches)):
                    if row[2][i][0]>0:
                        kmer = response["base"]+branches[i]
                        ckmer = haplotyping.General.canonical(kmer)
                        ckmerRow = ckmerTable[row[2][i][1]]
                        response["ckmers"][ckmer] = {
                            "number": int(ckmerRow[2]), 
                            "split": Split._translate_type(ckmerRow[1].decode("ascii"))
                        }
        else:
            response = None
        return response
                
    def _kmer_result(row, h5file: h5py.File, expandBase=True):
        if row:
            response = {
                "ckmer": row[0].decode("ascii"),
                "split": Split._translate_type(row[1].decode("ascii")),
                "number": int(row[2]),
                "rightSplitBase": {
                },
                "direct": {
                    "left": {
                        "distinct": int(row[4][1][0]),
                        "number": int(row[4][1][1])
                    },"right": {
                        "distinct": int(row[4][2][0]),
                        "number": int(row[4][2][1])
                    }
                },
                "cycle": int(row[6][0]),
                "reverse": int(row[7][0]),
                "paired": int(row[8][1]),                
            }
            baseTable = h5file.get("/split/base")
            if response["split"] in ["left","both"]:
                baseRow=baseTable[row[3][0]]
                if expandBase:
                    response["rightSplitBase"]["left"] = Split._base_result(baseRow,h5file,False)
                else:
                    response["rightSplitBase"]["left"] = baseRow[0].decode("ascii")
            if response["split"] in ["right","both"]:
                baseRow=baseTable[row[3][1]]
                if expandBase:
                    response["rightSplitBase"]["right"] = Split._base_result(baseRow,h5file,False)
                else:
                    response["rightSplitBase"]["right"] = baseRow[0].decode("ascii")            
        else:
            response = None
        return response
    
    def _kmer_direct_result(ckmerRow, directRows, h5file: h5py.File, expandBase=True):
        if (not ckmerRow==None) or (len(directRows)==0):
            response = {
                "ckmer": ckmerRow[0].decode("ascii"),
                "split": Split._translate_type(ckmerRow[1].decode("ascii")),
                "number": int(ckmerRow[2]),
                "direct": {
                    "left": [],
                    "right": []
                }
            } 
            ckmerTable = h5file.get("/split/ckmer")
            baseTable = h5file.get("/split/base")
            for directRow in directRows:
                fromDirection = directRow[0][1].decode("ascii")
                ckmerRow = ckmerTable[directRow[1][0]]
                item = {
                    "connection": {                    
                        "direction": Split._translate_type(directRow[1][1].decode("ascii")),
                        "distance": int(directRow[3]),
                        "number": int(directRow[2]),
                        "problem": int(directRow[4])
                    },
                    "ckmer": ckmerRow[0].decode("ascii"),
                    "split": Split._translate_type(ckmerRow[1].decode("ascii")),
                    "number": int(ckmerRow[2]),
                    "rightSplitBase": {
                    },                
                }
                if item["split"] in ["left","both"]:
                    baseRow=baseTable[ckmerRow[3][0]]
                    item["rightSplitBase"]["left"] = Split._base_result(baseRow,h5file,False)
                if item["split"] in ["right","both"]:
                    baseRow=baseTable[ckmerRow[3][1]]
                    item["rightSplitBase"]["right"] = Split._base_result(baseRow,h5file,False)
                if fromDirection=="l":
                    response["direct"]["left"].append(item)
                elif fromDirection=="r":
                    response["direct"]["right"].append(item)
        else:
            response = None
        return response
    
    def _kmer_read_result(kmerIds,readInfoList,readData,h5file,kmerDict={}):
        if len(kmerIds)>0:
            ckmerTable = h5file.get("/split/ckmer")
            response = []
            n = 0
            checkIds = set(kmerIds)
            for item in readInfoList:
                read = readData[n:n+item[0]]
                if len(checkIds.intersection(read))>0:
                    kmerList = []
                    for kmerId in read:
                        if not kmerId in kmerDict:
                            entry = ckmerTable[kmerId]
                            kmerDict[kmerId] = (entry[0].decode("ascii"),Split._translate_type(entry[1].decode("ascii")),int(entry[2]))
                        kmerList.append(kmerDict[kmerId])
                    response.append({"kmers": kmerList, "number": int(item[1])})
                n+=item[0]
        else:
            response = []
        return response,kmerDict
    
    def _kmer_paired_result(kmerId,pairedList,h5file,kmerDict={}):
        if len(pairedList)>0:
            ckmerTable = h5file.get("/split/ckmer")
            response = []
            for item in pairedList:
                if item[0]==kmerId:
                    if not item[1] in kmerDict:
                        kmerDict[item[1]] = ckmerTable[item[1]][0].decode("ascii")
                    response.append(kmerDict[item[1]])
        else:
            response = []
        return response,kmerDict
    
    #---
    
    def _info(h5file: h5py.File):
        response = {}
        configTable = h5file.get("/config")
        for k in configTable.attrs.keys():
            value = configTable.attrs[k]
            if np.issubdtype(type(value), np.integer):
                response[k] = int(value)
            else:
                response[k] = str(value)
        return response

    def _kmer_distribution(h5file: h5py.File):
        histogramTable = h5file.get("/histogram/kmer")
        configTable = h5file.get("/config")
        response = {}
        response["k"] = int(configTable.attrs["k"])
        response["minimumKmerFrequencies"] = int(configTable.attrs["minimumKmerFrequencies"])
        response["maximumKmerFrequencies"] = int(configTable.attrs["maximumKmerFrequencies"])
        response["numberKmers"] = int(configTable.attrs["numberKmers"])
        response["totalKmerFrequencies"] = int(configTable.attrs["totalKmerFrequencies"])
        response["frequencies"] = {int(item[0]): int(item[1]) for item in histogramTable}
        return response

    def _kmer_split_distribution(h5file: h5py.File):
        histogramTable = h5file.get("/histogram/ckmer")
        configTable = h5file.get("/config")
        response = {}
        response["k"] = int(configTable.attrs["k"])
        response["minimumCanonicalSplitFrequency"] = int(configTable.attrs["minimumCanonicalSplitFrequency"])
        response["maximumCanonicalSplitFrequency"] = int(configTable.attrs["maximumCanonicalSplitFrequency"])
        response["numberCanonicalSplit"] = int(configTable.attrs["numberCanonicalSplit"])
        response["numberCanonicalSplitBoth"] = int(configTable.attrs["numberCanonicalSplitBoth"])
        response["numberCanonicalSplitLeft"] = int(configTable.attrs["numberCanonicalSplitLeft"])
        response["numberCanonicalSplitRight"] = int(configTable.attrs["numberCanonicalSplitRight"])
        response["totalCanonicalSplitFrequencies"] = int(configTable.attrs["totalCanonicalSplitFrequencies"])
        response["frequencies"] = {int(item[0]): int(item[1]) for item in histogramTable}
        return response

    def _kmer_base_distribution(h5file: h5py.File):
        histogramTable = h5file.get("/histogram/base")
        configTable = h5file.get("/config")
        response = {}
        response["k"] = int(configTable.attrs["k"])
        response["numberRightSplitBases"] = int(configTable.attrs["numberRightSplitBases"])
        response["numberRightSplitKmers"] = int(configTable.attrs["numberRightSplitKmers"])
        response["frequencies"] = {int(item[0]): int(item[1]) for item in histogramTable}
        return response

    def _kmer_split_direct_distribution(h5file: h5py.File):
        histogramTable = h5file.get("/histogram/distance")
        configTable = h5file.get("/config")
        response = {}
        response["k"] = int(configTable.attrs["k"])
        response["maximumCycleLength"] = int(configTable.attrs["maximumCycleLength"])
        response["maximumReversalLength"] = int(configTable.attrs["maximumReversalLength"])
        response["numberCycles"] = int(configTable.attrs["numberCycles"])
        response["numberReversals"] = int(configTable.attrs["numberReversals"])
        response["frequencies"] = {int(item[0]): int(item[1]) for item in histogramTable}
        return response
    
    def _kmer_info(h5file: h5py.File, kmer: str):
        ckmer = haplotyping.General.canonical(kmer)
        ckmerTable = h5file.get("/split/ckmer")
        (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable)
        return Split._kmer_result(ckmerRow,h5file)
    
    def _kmers_info(h5file: h5py.File, kmers: list):
        ckmerList = set()
        for kmer in kmers:
            ckmerList.add(haplotyping.General.canonical(kmer))
        ckmerList = list(ckmerList)
        ckmerList.sort()
        response = []
        ckmerTable = h5file.get("/split/ckmer")
        number = ckmerTable.shape[0]
        start = 0
        cache = {}
        for i in range(len(ckmerList)):
            ckmer = ckmerList[i]
            (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable,start,number,cache)
            if ckmerRow:
                response.append(Split._kmer_result(ckmerRow,h5file))
                start = id+1
            else:
                start = id 
        return response
    
    def _kmers_connected(h5file: h5py.File, kmers: list):
        ckmerList = set()
        for kmer in kmers:
            ckmerList.add(haplotyping.General.canonical(kmer))
        ckmerList = list(ckmerList)
        ckmerList.sort()
        response = []
        partitions = set()
        kmerIds = set()
        ckmerTable = h5file.get("/split/ckmer")
        readPartitionTable = h5file.get("/relations/readPartition")
        readDataTable = h5file.get("/relations/readData")
        readInfoTable = h5file.get("/relations/readInfo")
        #get partitions
        number = ckmerTable.shape[0]
        start = 0
        cache = {}
        for i in range(len(ckmerList)):
            ckmer = ckmerList[i]
            (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable,start,number,cache)
            if ckmerRow:
                partitions.add(int(ckmerRow[5]))
                kmerIds.add(id)
                start = id+1
            else:
                start = id 
        #check reads from partitions
        partitions = list(partitions)
        partitions.sort()
        for i in range(len(partitions)):
            partition = readPartitionTable[partitions[i]]
            readData = readDataTable[partition[0][0]:partition[0][0]+partition[0][1]]
            readInfo = readInfoTable[partition[1][0]:partition[1][0]+partition[1][1]]
            
            print(len(readData),"readData")
            for item in readData:
                print(item)
            print(len(readInfo),"readInfo")
            for item in readInfo:
                print(item)
            print(partition)
            print(readPartitionTable[partitions[i]+1])
        return response
    
    def _kmer_direct(h5file: h5py.File, kmer: str):
        ckmer = haplotyping.General.canonical(kmer)
        ckmerTable = h5file.get("/split/ckmer")
        (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable)
        if ckmerRow:
            directTable = h5file.get("/relations/direct")
            directRows = directTable[ckmerRow[4][0]:ckmerRow[4][0]+(ckmerRow[4][1][0]+ckmerRow[4][2][0])]
            return Split._kmer_direct_result(ckmerRow,directRows,h5file)
        else:
            return None
    
    def _kmers_direct(h5file: h5py.File, kmers: list):
        ckmerList = set()
        for kmer in kmers:
            ckmerList.add(haplotyping.General.canonical(kmer))
        ckmerList = list(ckmerList)
        ckmerList.sort()
        response = []
        ckmerTable = h5file.get("/split/ckmer")
        directTable = h5file.get("/relations/direct")
        number = ckmerTable.shape[0]
        start = 0
        cache = {}
        for i in range(len(ckmerList)):
            ckmer = ckmerList[i]
            (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable,start,number,cache)
            if ckmerRow:
                directRows = directTable[ckmerRow[4][0]:ckmerRow[4][0]+(ckmerRow[4][1][0]+ckmerRow[4][2][0])]
                response.append(Split._kmer_direct_result(ckmerRow,directRows,h5file))
                start = id+1
            else:
                start = id 
        return list(filter(None, response))
   
    def _kmer_read(h5file: h5py.File, kmer: str):
        ckmer = haplotyping.General.canonical(kmer)
        ckmerTable = h5file.get("/split/ckmer")
        readPartitionTable = h5file.get("/relations/readPartition")
        readDataTable = h5file.get("/relations/readData")
        readInfoTable = h5file.get("/relations/readInfo")
        (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable)
        if ckmerRow:
            kmerDict = {id: (ckmerRow[0].decode("ascii"),Split._translate_type(ckmerRow[1].decode("ascii")),int(ckmerRow[2]))}
            partitionRow = readPartitionTable[ckmerRow[5]]
            readDataList = readDataTable[partitionRow[0][0]:partitionRow[0][0]+partitionRow[0][1]]
            readInfoList = readInfoTable[partitionRow[1][0]:partitionRow[1][0]+partitionRow[1][1]]
            reads,kmerDict = Split._kmer_read_result([id],readInfoList,readDataList,h5file,kmerDict)
        else:
            reads = []
        return reads
        
    def _kmers_read(h5file: h5py.File, kmers: list):
        ckmerList = set()
        for kmer in kmers:
            ckmerList.add(haplotyping.General.canonical(kmer))
        ckmerList = list(ckmerList)
        ckmerList.sort()
        ckmerTable = h5file.get("/split/ckmer")
        readPartitionTable = h5file.get("/relations/readPartition")
        readDataTable = h5file.get("/relations/readData")
        readInfoTable = h5file.get("/relations/readInfo")
        number = ckmerTable.shape[0]
        start = 0
        cache = {}
        partitions = set()
        kmerIds = []
        kmerDict = {}
        #get partition data from k-mers
        for i in range(len(ckmerList)):
            ckmer = ckmerList[i]
            (ckmerRow,id,cache) = Split._findItem(ckmerList[i],ckmerTable,start,number,cache)
            if ckmerRow:
                kmerDict[id] = (ckmerRow[0].decode("ascii"),Split._translate_type(ckmerRow[1].decode("ascii")),int(ckmerRow[2]))
                kmerIds.append(id)
                partitions.add(ckmerRow[5])
                start = id+1
            else:
                start = id 
        #collect reads from partitions
        reads = []
        for p in partitions:
            partitionRow = readPartitionTable[p]
            readDataList = readDataTable[partitionRow[0][0]:partitionRow[0][0]+partitionRow[0][1]]
            readInfoList = readInfoTable[partitionRow[1][0]:partitionRow[1][0]+partitionRow[1][1]]
            newReads,kmerDict = Split._kmer_read_result(kmerIds,readInfoList,readDataList,h5file,kmerDict)
            reads.extend(newReads)
        return reads
    
    def _kmer_paired(h5file: h5py.File, kmer: str):
        ckmer = haplotyping.General.canonical(kmer)
        ckmerTable = h5file.get("/split/ckmer")
        pairedTable = h5file.get("/relations/paired")
        (ckmerRow,id,cache) = Split._findItem(ckmer,ckmerTable)
        if ckmerRow:
            kmerDict = {id: ckmerRow[0].decode("ascii")}
            pairedList = pairedTable[ckmerRow[8][0]:ckmerRow[8][0]+ckmerRow[8][1]]
            paired,kmerDict = Split._kmer_paired_result(id,pairedList,h5file,kmerDict)
        else:
            paired = []
        return paired
        
    def _kmers_paired(h5file: h5py.File, kmers: list):
        ckmerList = set()
        for kmer in kmers:
            ckmerList.add(haplotyping.General.canonical(kmer))
        ckmerList = list(ckmerList)
        ckmerList.sort()
        ckmerTable = h5file.get("/split/ckmer")
        pairedTable = h5file.get("/relations/paired")
        number = ckmerTable.shape[0]
        start = 0
        cache = {}
        kmerDict = {}
        response = {}
        #get paired data for k-mers
        for i in range(len(ckmerList)):
            ckmer = ckmerList[i]
            (ckmerRow,id,cache) = Split._findItem(ckmerList[i],ckmerTable,start,number,cache)
            if ckmerRow:
                kmerDict[id] = ckmerRow[0].decode("ascii")
                if ckmerRow[8][1]>0:
                    pairedList = pairedTable[ckmerRow[8][0]:ckmerRow[8][0]+ckmerRow[8][1]]
                    response[kmerDict[id]],kmerDict = Split._kmer_paired_result(id,pairedList,h5file,kmerDict)
                start = id+1
            else:
                start = id 
        return response
    
    #---
    
    def info(location_split: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._info(h5file)
    
   #---
    
    def kmer_distribution(location_split: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_distribution(h5file)

    def kmer_split_distribution(location_split: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_split_distribution(h5file)

    def kmer_base_distribution(location_split: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_base_distribution(h5file)

    def kmer_split_direct_distribution(location_split: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_split_direct_distribution(h5file)
    
   #---
    
    def kmer_info(location_split: str, kmer: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_info(h5file,kmer)
    
    def kmer_list_info(location_split: str, kmers: list):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmers_info(h5file,kmers)
    
    def kmer_sequence_info(location_split: str, sequence: str):
        with h5py.File(location_split, mode="r") as h5file:            
            configTable = h5file.get("/config")
            k = int(configTable.attrs["k"])
            kmers = [sequence[i:i+k] for i in range(len(sequence)-(k-1))] 
            return Split._kmers_info(h5file,kmers)
        
    def kmer_direct(location_split: str, kmer: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_direct(h5file,kmer)
    
    def kmer_list_direct(location_split: str, kmers: list):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmers_direct(h5file,kmers)
        
    def kmer_read(location_split: str, kmer: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_read(h5file,kmer)
    
    def kmer_list_read(location_split: str, kmers: list):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmers_read(h5file,kmers)
        
    def kmer_paired(location_split: str, kmer: str):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmer_paired(h5file,kmer)
    
    def kmer_list_paired(location_split: str, kmers: list):
        with h5py.File(location_split, mode="r") as h5file:            
            return Split._kmers_paired(h5file,kmers)
        
    #---
    
    def base_info(location_split: str, base: str):
        with h5py.File(location_split, mode="r") as h5file:            
            baseTable = h5file.get("/split/base")
            (baseRow,id,cache) = Split._findItem(base,baseTable)
            return Split._base_result(baseRow,h5file)
        
    def base_list_info(location_split: str, bases: list):
        baseList = []
        for base in set(bases):
            baseList.append(base)
        baseList.sort()
        response = []
        with h5py.File(location_split, mode="r") as h5file:            
            baseTable = h5file.get("/split/base")
            number = baseTable.shape[0]
            start = 0
            cache = {}
            for i in range(len(baseList)):
                base = baseList[i]
                (baseRow,id,cache) = Split._findItem(base,baseTable,start,number,cache)
                if not baseRow==None:
                    response.append(Split._base_result(baseRow,h5file))
                    start = id+1
                else:
                    start = id 
        return response
        
        


from flask import Response, request, abort
from flask_restx import Namespace, Resource
import json, haplotyping, sqlite3

namespace = Namespace("collection", description="Datasets all belong to a collection", path="/collection")

def adjust_collection_response(item):
    item["datasets"] = {
        "total": item["datasets_total"], 
        "kmer": item["datasets_kmer"],
        "split": item["datasets_split"],
        "marker": item["datasets_marker"]
    }
    del item["datasets_total"]
    del item["datasets_kmer"]
    del item["datasets_split"]
    del item["datasets_marker"]
    #finished
    return item


@namespace.route("/")
class CollectionList(Resource):
    
    @namespace.doc(description="Get collections")
    @namespace.param("start", "start of listed results", type="int", required=False)
    @namespace.param("number", "(maximum) number of listed results", type="int", required=False)
    def get(self):
        try:
            start = int(request.args.get("start",0))
            number = int(request.args.get("number",10))        
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("SELECT COUNT(*) AS `number` FROM `collection`")  
            total = cursor.fetchone()[0]
            if start<total:
                cursor.execute("""SELECT `collection`.`uid`, `collection`.`name`,
                                       `collection`.`experiment`, `collection`.`type`,
                                       COUNT(DISTINCT(`dataset`.`id`)) AS `datasets`,
                                       COUNT(DISTINCT(`variety`.`id`)) AS `varieties`
                                FROM `collection`
                                INNER JOIN `dataset` ON `collection`.`id` = `dataset`.`collection_id`
                                LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid`
                                WHERE NOT `dataset`.`uid` IS NULL AND NOT `dataset`.`type` IS NULL
                                GROUP BY `collection`.`id` ORDER BY `collection`.`name`,
                                `collection`.`type`, `collection`.`id` LIMIT ?,?""",(start,number,))  
                resultList = [dict(row) for row in cursor.fetchall()]
            else:
                resultList = []
            response = {"start": start, "number": number, "total": total, "list": resultList}
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>")
@namespace.doc(params={"uid": "unique identifier collection"})
class CollectionId(Resource):
    
    @namespace.doc(description="Get collection by uid")
    def get(self,uid):
        try:
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("""SELECT `collection`.`uid`, `collection`.`name`,
                                       `collection`.`experiment`, `collection`.`type`,
                                       COUNT(DISTINCT(`dataset`.`id`)) AS `datasets`,
                                       COUNT(DISTINCT(`variety`.`id`)) AS `varieties`
                                FROM `collection`
                                INNER JOIN `dataset` ON `collection`.`id` = `dataset`.`collection_id`
                                LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid`
                                WHERE NOT `dataset`.`uid` IS NULL AND NOT `dataset`.`type` IS NULL
                                AND `collection`.`uid` = ?
                                GROUP BY `collection`.`id`""",(uid,))  
            data = cursor.fetchone()
            if data:
                return Response(json.dumps(dict(data)), mimetype="application/json")
            else:
                abort(404, "no collection with uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))            

